"""
This module defines the `VariableOwner` class, a mix-in class for everything
that saves state variables, e.g. `Clock` or `NeuronGroup`, the class `Group`
for objects that in addition to storing state variables also execute code, i.e.
objects such as `NeuronGroup` or `StateMonitor` but not `Clock`, and finally
`CodeRunner`, a class to run code in the context of a `Group`.
"""

import inspect
import numbers
import weakref
from collections import OrderedDict
from collections.abc import Mapping

import numpy as np

from brian2.codegen.codeobject import create_runner_codeobj
from brian2.core.base import BrianObject, weakproxy_with_fallback
from brian2.core.functions import Function
from brian2.core.names import Nameable, find_name
from brian2.core.namespace import (
    DEFAULT_CONSTANTS,
    DEFAULT_FUNCTIONS,
    DEFAULT_UNITS,
    get_local_namespace,
)
from brian2.core.preferences import prefs
from brian2.core.variables import (
    ArrayVariable,
    AuxiliaryVariable,
    Constant,
    DynamicArrayVariable,
    LinkedVariable,
    Subexpression,
    Variable,
    Variables,
)
from brian2.equations.equations import BOOLEAN, FLOAT, INTEGER, Equations
from brian2.importexport.importexport import ImportExport
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    DimensionMismatchError,
    fail_for_dimension_mismatch,
    get_unit,
)
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import SpellChecker, get_identifiers

__all__ = ["Group", "VariableOwner", "CodeRunner"]

logger = get_logger(__name__)


def _display_value(obj):
    """
    Helper function for warning messages that display the value of objects. This
    functions returns a nicer representation for symbolic constants and
    functions.

    Parameters
    ----------
    obj : object
        The object to display

    Returns
    -------
    value : str
        A string representation of the object
    """
    if isinstance(obj, Function):
        return "<Function>"
    try:
        obj = obj.get_value()
    except AttributeError:
        pass
    try:
        obj = obj.value
    except AttributeError:
        pass

    # We (temporarily) set numpy's print options so that array with more than
    # 10 elements are only shown in an abbreviated way
    old_options = np.get_printoptions()
    np.set_printoptions(threshold=10)
    try:
        str_repr = repr(obj)
    except Exception:
        str_repr = f"<object of type {type(obj)}>"
    finally:
        np.set_printoptions(**old_options)
    return str_repr


def _conflict_warning(message, resolutions):
    """
    A little helper functions to generate warnings for logging. Specific
    to the `Group._resolve` method and should only be used by it.

    Parameters
    ----------
    message : str
        The first part of the warning message.
    resolutions : list of str
        A list of (namespace, object) tuples.
    """
    if len(resolutions) == 0:
        # nothing to warn about
        return
    elif len(resolutions) == 1:
        second_part = (
            "but the name also refers to a variable in the "
            f"{resolutions[0][0]} "
            f"namespace with value '{_display_value(resolutions[0][1])}'."
        )
    else:
        second_part = (
            "but the name also refers to a variable in the following "
            f"namespaces: {', '.join([r[0] for r in resolutions])}."
        )

    logger.warn(
        f"{message} {second_part}", "Group.resolve.resolution_conflict", once=True
    )


def get_dtype(equation, dtype=None):
    """
    Helper function to interpret the `dtype` keyword argument in `NeuronGroup`
    etc.

    Parameters
    ----------
    equation : `SingleEquation`
        The equation for which a dtype should be returned
    dtype : `dtype` or dict, optional
        Either the `dtype` to be used as a default dtype for all float variables
        (instead of the `core.default_float_dtype` preference) or a
        dictionary stating the `dtype` for some variables; all other variables
        will use the preference default

    Returns
    -------
    d : `dtype`
        The dtype for the variable defined in `equation`
    """
    # Check explicitly provided dtype for compatibility with the variable type
    if isinstance(dtype, Mapping):
        if equation.varname in dtype:
            BASIC_TYPES = {BOOLEAN: "b", INTEGER: "iu", FLOAT: "f"}
            provided_dtype = np.dtype(dtype[equation.varname])
            if provided_dtype.kind not in BASIC_TYPES[equation.var_type]:
                raise TypeError(
                    "Error determining dtype for variable "
                    f"{equation.varname}: {provided_dtype.name} is not "
                    f"a correct type for {equation.var_type} variables"
                )
            else:
                return dtype[equation.varname]
        else:  # continue as if no dtype had been specified at all
            dtype = None

    # Use default dtypes (or a provided standard dtype for floats)
    if equation.var_type == BOOLEAN:
        return bool
    elif equation.var_type == INTEGER:
        return prefs["core.default_integer_dtype"]
    elif equation.var_type == FLOAT:
        if dtype is not None:
            dtype = np.dtype(dtype)
            if not dtype.kind == "f":
                raise TypeError(f"{dtype} is not a valid floating point dtype.")
            return dtype
        else:
            return prefs["core.default_float_dtype"]
    else:
        raise ValueError(
            "Do not know how to determine a dtype for "
            f"variable '{equation.varname}' of type {equation.var_type}"
        )


def _same_value(obj1, obj2):
    """
    Helper function used during namespace resolution.
    """
    if obj1 is obj2:
        return True
    try:
        obj1 = obj1.get_value()
    except (AttributeError, TypeError):
        pass

    try:
        obj2 = obj2.get_value()
    except (AttributeError, TypeError):
        pass

    return obj1 is obj2


def _same_function(func1, func2):
    """
    Helper function, used during namespace resolution for comparing whether to
    functions are the same. This takes care of treating a function and a
    `Function` variables whose `Function.pyfunc` attribute matches as the
    same. This prevents the user from getting spurious warnings when having
    for example a numpy function such as :np:func:`~random.randn` in the local
    namespace, while the ``randn`` symbol in the numpy namespace used for the
    code objects refers to a `RandnFunction` specifier.
    """
    # use the function itself if it doesn't have a pyfunc attribute
    func1 = getattr(func1, "pyfunc", func1)
    func2 = getattr(func2, "pyfunc", func2)

    return func1 is func2


class Indexing:
    """
    Object responsible for calculating flat index arrays from arbitrary group-
    specific indices. Stores strong references to the necessary variables so
    that basic indexing (i.e. slicing, integer arrays/values, ...) works even
    when the respective `VariableOwner` no longer exists. Note that this object
    does not handle string indexing.
    """

    def __init__(self, group, default_index="_idx"):
        self.group = weakref.proxy(group)
        self.N = group.variables["N"]
        self.default_index = default_index

    def __call__(self, item=slice(None), index_var=None):  # noqa: B008
        """
        Return flat indices to index into state variables from arbitrary
        group specific indices. In the default implementation, raises an error
        for multidimensional indices and transforms slices into arrays.

        Parameters
        ----------
        item : slice, array, int
            The indices to translate.

        Returns
        -------
        indices : `numpy.ndarray`
            The flat indices corresponding to the indices given in `item`.

        See Also
        --------
        SynapticIndexing
        """
        if index_var is None:
            index_var = self.default_index

        if hasattr(item, "_indices"):
            item = item._indices()

        if isinstance(item, tuple):
            raise IndexError(
                f"Can only interpret 1-d indices, got {len(item)} dimensions."
            )
        else:
            if isinstance(item, str) and item == "True":
                item = slice(None)
            if isinstance(item, slice):
                if index_var == "0":
                    return 0
                if index_var == "_idx":
                    start, stop, step = item.indices(self.N.item())
                else:
                    start, stop, step = item.indices(index_var.size)
                index_array = np.arange(start, stop, step, dtype=np.int32)
            else:
                index_array = np.asarray(item)
                if index_array.dtype == bool:
                    index_array = np.nonzero(index_array)[0]
                elif not np.issubdtype(index_array.dtype, np.signedinteger):
                    raise TypeError(
                        "Indexing is only supported for integer "
                        "and boolean arrays, not for type "
                        f"{index_array.dtype}"
                    )

            if index_var != "_idx":
                try:
                    return index_var.get_value()[index_array]
                except IndexError as ex:
                    # We try to emulate numpy's indexing semantics here:
                    # slices never lead to IndexErrors, instead they return an
                    # empty array if they don't match anything
                    if isinstance(item, slice):
                        return np.array([], dtype=np.int32)
                    else:
                        raise ex
            else:
                return index_array


class IndexWrapper:
    """
    Convenience class to allow access to the indices via indexing syntax. This
    allows for example to get all indices for synapses originating from neuron
    10 by writing `synapses.indices[10, :]` instead of
    `synapses._indices.((10, slice(None))`.
    """

    def __init__(self, group):
        self.group = weakref.proxy(group)
        self.indices = group._indices

    def __getitem__(self, item):
        if isinstance(item, str):
            variables = Variables(None)
            variables.add_auxiliary_variable("_indices", dtype=np.int32)
            variables.add_auxiliary_variable("_cond", dtype=bool)

            abstract_code = f"_cond = {item}"
            namespace = get_local_namespace(level=1)
            from brian2.devices.device import get_device

            device = get_device()
            codeobj = create_runner_codeobj(
                self.group,
                abstract_code,
                "group_get_indices",
                run_namespace=namespace,
                additional_variables=variables,
                codeobj_class=device.code_object_class(
                    fallback_pref="codegen.string_expression_target"
                ),
            )
            return codeobj()
        else:
            return self.indices(item)


class VariableOwner(Nameable):
    """
    Mix-in class for accessing arrays by attribute.

    # TODO: Overwrite the __dir__ method to return the state variables
    # (should make autocompletion work)
    """

    def _enable_group_attributes(self):
        if not hasattr(self, "variables"):
            raise ValueError(
                "Classes derived from VariableOwner need a variables attribute."
            )
        if "N" not in self.variables:
            raise ValueError("Each VariableOwner needs an 'N' variable.")
        if not hasattr(self, "codeobj_class"):
            self.codeobj_class = None
        if not hasattr(self, "_indices"):
            self._indices = Indexing(self)
        if not hasattr(self, "indices"):
            self.indices = IndexWrapper(self)
        if not hasattr(self, "_stored_states"):
            self._stored_states = {}
        self._group_attribute_access_active = True

    def state(self, name, use_units=True, level=0):
        """
        Return the state variable in a way that properly supports indexing in
        the context of this group

        Parameters
        ----------
        name : str
            The name of the state variable
        use_units : bool, optional
            Whether to use the state variable's unit.
        level : int, optional
            How much farther to go down in the stack to find the namespace.
        Returns
        -------
        var : `VariableView` or scalar value
            The state variable's value that can be indexed (for non-scalar
            values).
        """
        try:
            var = self.variables[name]
        except KeyError as exc:
            raise KeyError(f"State variable {name} not found.") from exc

        if use_units:
            return var.get_addressable_value_with_unit(name=name, group=self)
        else:
            return var.get_addressable_value(name=name, group=self)

    def __getattr__(self, name):
        # We do this because __setattr__ and __getattr__ are not active until
        # _group_attribute_access_active attribute is set, and if it is set,
        # then __getattr__ will not be called. Therefore, if getattr is called
        # with this name, it is because it hasn't been set yet and so this
        # method should raise an AttributeError to agree that it hasn't been
        # called yet.
        if name == "_group_attribute_access_active":
            raise AttributeError
        if "_group_attribute_access_active" not in self.__dict__:
            raise AttributeError
        if (
            name in self.__getattribute__("__dict__")
            or name in self.__getattribute__("__class__").__dict__
        ):
            # Makes sure that classes can override the "variables" mechanism
            # with instance/class attributes and properties
            return object.__getattribute__(self, name)
        # We want to make sure that accessing variables without units is fast
        # because this is what is used during simulations
        # We do not specifically check for len(name) here, we simply assume
        # that __getattr__ is not called with an empty string (which wouldn't
        # be possible using the normal dot syntax, anyway)
        try:
            if name[-1] == "_":
                name = name[:-1]
                use_units = False
            else:
                use_units = True
            return self.state(name, use_units)

        except KeyError:
            raise AttributeError(f"No attribute with name {name}")

    def __setattr__(self, key, value, level=0):
        # attribute access is switched off until this attribute is created by
        # _enable_group_attributes
        if not hasattr(self, "_group_attribute_access_active") or key in self.__dict__:
            object.__setattr__(self, key, value)
        elif key in getattr(self, "_linked_variables", set()):
            if not isinstance(value, LinkedVariable):
                raise ValueError(
                    "Cannot set a linked variable directly, link "
                    "it to another variable using 'linked_var'."
                )
            linked_var = value.variable

            eq = self.equations[key]
            if eq.dim is not linked_var.dim:
                raise DimensionMismatchError(
                    f"Unit of variable '{key}' does not "
                    "match its link target "
                    f"'{linked_var.name}'"
                )

            if not isinstance(linked_var, Subexpression):
                var_length = len(linked_var)
            else:
                var_length = len(linked_var.owner)

            if value.index is not None:
                index = self._linked_var_index(key, value, var_length)
            else:
                index = self._linked_var_automatic_index(key, value, var_length)
            self.variables.add_reference(key, value.group, value.name, index=index)
            source = (value.variable.owner.name,)
            sourcevar = value.variable.name
            log_msg = f"Setting {self.name}.{key} as a link to {source}.{sourcevar}"
            if index is not None:
                log_msg += f'(using "{index}" as index variable)'
            logger.diagnostic(log_msg)
        else:
            if isinstance(value, LinkedVariable):
                raise TypeError(
                    f"Cannot link variable '{key}', it has to be marked "
                    "as a linked variable with '(linked)' in the model "
                    "equations."
                )
            else:
                if (
                    key in self.__getattribute__("__dict__")
                    or key in self.__getattribute__("__class__").__dict__
                ):
                    # Makes sure that classes can override the "variables" mechanism
                    # with instance/class attributes and properties
                    return object.__setattr__(self, key, value)
                elif key in self.variables:
                    var = self.variables[key]
                    if not isinstance(value, str):
                        if var.dim is DIMENSIONLESS:
                            fail_for_dimension_mismatch(
                                value,
                                var.dim,
                                "%s should be set with a dimensionless value, but got {value}"
                                % key,
                                value=value,
                            )
                        else:
                            fail_for_dimension_mismatch(
                                value,
                                var.dim,
                                "%s should be set with a value with units %r, but got {value}"
                                % (key, get_unit(var.dim)),
                                value=value,
                            )
                    if var.read_only:
                        raise TypeError(f"Variable {key} is read-only.")
                    # Make the call X.var = ... equivalent to X.var[:] = ...
                    var.get_addressable_value_with_unit(key, self).set_item(
                        slice(None), value, level=level + 1
                    )
                elif len(key) and key[-1] == "_" and key[:-1] in self.variables:
                    # no unit checking
                    var = self.variables[key[:-1]]
                    if var.read_only:
                        raise TypeError(f"Variable {key[:-1]} is read-only.")
                    # Make the call X.var = ... equivalent to X.var[:] = ...
                    var.get_addressable_value(key[:-1], self).set_item(
                        slice(None), value, level=level + 1
                    )
                elif hasattr(self, key) or key.startswith("_"):
                    object.__setattr__(self, key, value)
                else:
                    # Try to suggest the correct name in case of a typo
                    checker = SpellChecker(
                        [
                            varname
                            for varname, var in self.variables.items()
                            if not (varname.startswith("_") or var.read_only)
                        ]
                    )
                    if key.endswith("_"):
                        suffix = "_"
                        key = key[:-1]
                    else:
                        suffix = ""
                    error_msg = f'Could not find a state variable with name "{key}".'
                    suggestions = checker.suggest(key)
                    if len(suggestions) == 1:
                        (suggestion,) = suggestions
                        error_msg += f' Did you mean to write "{suggestion}{suffix}"?'
                    elif len(suggestions) > 1:
                        suggestion_str = ", ".join(
                            [f"'{suggestion}{suffix}'" for suggestion in suggestions]
                        )
                        error_msg += f" Did you mean to write any of the following: {suggestion_str} ?"
                    error_msg += (
                        " Use the add_attribute method if you intend to add "
                        "a new attribute to the object."
                    )
                    raise AttributeError(error_msg)

    def _linked_var_automatic_index(self, var_name, linked_var, var_length):
        # The check at the end is to avoid the case that a size 1 NeuronGroup
        # links to another NeuronGroup of size 1 and cannot do certain operations
        # since the linked variable is considered scalar.
        if linked_var.variable.scalar or (
            var_length == 1 and getattr(self, "_N", 0) != 1
        ):
            index = "0"
        else:
            index = linked_var.group.variables.indices[linked_var.name]
            if index == "_idx":
                target_length = var_length
            else:
                target_length = len(linked_var.group.variables[index])
                # we need a name for the index that does not clash with
                # other names and a reference to the index
                new_index = f"_{linked_var.name}_index_{index}"
                self.variables.add_reference(new_index, linked_var.group, index)
                index = new_index

            if len(self) != target_length:
                raise ValueError(
                    f"Cannot link variable '{var_name}' to "
                    f"'{linked_var.variable.name}', the size of the "
                    "target group does not match "
                    f"({len(self)} != {target_length}). You can "
                    "provide an indexing scheme with the "
                    "'index' keyword to link groups with "
                    "different sizes"
                )
        return index

    def _linked_var_index(self, var_name, linked_var, target_size):
        if isinstance(linked_var.index, str):
            if linked_var.index not in self.variables:
                raise ValueError(f"Index variable '{linked_var.index}' not found.")
            if self.variables.indices[linked_var.index] != self.variables.default_index:
                raise ValueError(
                    f"Index variable '{linked_var.index}' should use the default index itself."
                )
            if not np.issubdtype(self.variables[linked_var.index].dtype, np.integer):
                raise TypeError(
                    f"Index variable '{linked_var.index}' should be an integer parameter."
                )
            index = linked_var.index
        else:
            # Index arrays are not allowed for classes with dynamic size (Synapses)
            if not isinstance(self.variables["N"], Constant):
                raise TypeError(
                    "Cannot link a variable with an index array for a class with dynamic size – use a variable name instead."
                )
            try:
                index_array = np.asarray(linked_var.index)
                if not np.issubdtype(index_array.dtype, int):
                    raise TypeError()
            except TypeError:
                raise TypeError(
                    "The index for a linked variable has to be an integer array"
                )
            size = len(index_array)
            source_index = linked_var.group.variables.indices[linked_var.name]
            if source_index not in ("_idx", "0"):
                # we are indexing into an already indexed variable,
                # calculate the indexing into the target variable
                index_array = linked_var.group.variables[source_index].get_value()[
                    index_array
                ]

            if not index_array.ndim == 1 or size != len(self):
                raise TypeError(
                    f"Index array for linked variable '{var_name}' "
                    "has to be a one-dimensional array of "
                    f"length {len(self)}, but has shape "
                    f"{index_array.shape!s}"
                )
            if min(index_array) < 0 or max(index_array) >= target_size:
                raise ValueError(
                    f"Index array for linked variable {var_name} "
                    "contains values outside of the valid "
                    f"range [0, {target_size}["
                )
            self.variables.add_array(
                f"_{var_name}_indices",
                size=size,
                dtype=index_array.dtype,
                constant=True,
                read_only=True,
                values=index_array,
            )
            index = f"_{var_name}_indices"
        return index

    def add_attribute(self, name):
        """
        Add a new attribute to this group. Using this method instead of simply
        assigning to the new attribute name is necessary because Brian will
        raise an error in that case, to avoid bugs passing unnoticed
        (misspelled state variable name, un-declared state variable, ...).

        Parameters
        ----------
        name : str
            The name of the new attribute

        Raises
        ------
        AttributeError
            If the name already exists as an attribute or a state variable.
        """
        if name in self.variables:
            raise AttributeError(
                f'Cannot add an attribute "{name}", it is already a state variable of'
                " this group."
            )
        if hasattr(self, name):
            raise AttributeError(
                f'Cannot add an attribute "{name}", it is already an attribute of this'
                " group."
            )
        object.__setattr__(self, name, None)

    def get_states(
        self,
        vars=None,
        units=True,
        format="dict",
        subexpressions=False,
        read_only_variables=True,
        level=0,
    ):
        """
        Return a copy of the current state variable values. The returned arrays
        are copies of the actual arrays that store the state variable values,
        therefore changing the values in the returned dictionary will not affect
        the state variables.

        Parameters
        ----------
        vars : list of str, optional
            The names of the variables to extract. If not specified, extract
            all state variables (except for internal variables, i.e. names that
            start with ``'_'``). If the ``subexpressions`` argument is ``True``,
            the current values of all subexpressions are returned as well.
        units : bool, optional
            Whether to include the physical units in the return value. Defaults
            to ``True``.
        format : str, optional
            The output format. Defaults to ``'dict'``.
        subexpressions: bool, optional
            Whether to return subexpressions when no list of variable names
            is given. Defaults to ``False``. This argument is ignored if an
            explicit list of variable names is given in ``vars``.
        read_only_variables : bool, optional
            Whether to return read-only variables (e.g. the number of neurons,
            the time, etc.). Setting it to ``False`` will assure that the
            returned state can later be used with `set_states`. Defaults to
            ``True``.
        level : int, optional
            How much higher to go up the stack to resolve external variables.
            Only relevant if extracting subexpressions that refer to external
            variables.

        Returns
        -------
        values : dict or specified format
            The variables specified in ``vars``, in the specified ``format``.

        """
        if format not in ImportExport.methods:
            raise NotImplementedError(f"Format '{format}' is not supported")
        if vars is None:
            vars = []
            for name, var in self.variables.items():
                if name.startswith("_"):
                    continue
                if subexpressions or not isinstance(var, Subexpression):
                    if read_only_variables or not getattr(var, "read_only", False):
                        if not isinstance(var, AuxiliaryVariable):
                            vars.append(name)
        data = ImportExport.methods[format].export_data(
            self, vars, units=units, level=level
        )
        return data

    def set_states(self, values, units=True, format="dict", level=0):
        """
        Set the state variables.

        Parameters
        ----------
        values : depends on ``format``
            The values according to ``format``.
        units : bool, optional
            Whether the ``values`` include physical units. Defaults to ``True``.
        format : str, optional
            The format of ``values``. Defaults to ``'dict'``
        level : int, optional
            How much higher to go up the stack to resolve external variables.
            Only relevant when using string expressions to set values.
        """
        # For the moment, 'dict' is the only supported format -- later this will
        # be made into an extensible system, see github issue #306
        if format not in ImportExport.methods:
            raise NotImplementedError(f"Format '{format}' is not supported")
        ImportExport.methods[format].import_data(self, values, units=units, level=level)

    def check_variable_write(self, variable):
        """
        Function that can be overwritten to raise an error if writing to a
        variable should not be allowed. Note that this does *not* deal with
        incorrect writes that are general to all kind of variables (incorrect
        units, writing to a read-only variable, etc.). This function is only
        used for type-specific rules, e.g. for raising an error in `Synapses`
        when writing to a synaptic variable before any `~Synapses.connect`
        call.

        By default this function does nothing.

        Parameters
        ----------
        variable : `Variable`
            The variable that the user attempts to set.
        """
        pass

    def _full_state(self):
        state = {}
        for var in self.variables.values():
            if not isinstance(var, ArrayVariable):
                continue  # we are only interested in arrays
            if var.owner is None or var.owner.name != self.name:
                continue  # we only store the state of our own variables

            state[var.name] = (var.get_value().copy(), var.size)

        return state

    def _restore_from_full_state(self, state):
        for var_name, (values, size) in state.items():
            var = self.variables[var_name]
            if isinstance(var, DynamicArrayVariable):
                var.resize(size)
            var.set_value(values)

    def _check_expression_scalar(self, expr, varname, level=0, run_namespace=None):
        """
        Helper function to check that an expression only refers to scalar
        variables, used when setting a scalar variable with a string expression.

        Parameters
        ----------
        expr : str
            The expression to check.
        varname : str
            The variable that is being set (only used for the error message)
        level : int, optional
            How far to go up in the stack to find the local namespace (if
            `run_namespace` is not set).
        run_namespace : dict-like, optional
            A specific namespace provided for this expression.

        Raises
        ------
        ValueError
            If the expression refers to a non-scalar variable.
        """
        identifiers = get_identifiers(expr)
        referred_variables = self.resolve_all(
            identifiers, run_namespace=run_namespace, level=level + 1
        )
        for ref_varname, ref_var in referred_variables.items():
            if not getattr(ref_var, "scalar", False):
                raise ValueError(
                    "String expression for setting scalar "
                    f"variable '{varname} refers to '{ref_varname} which "
                    "is not scalar."
                )

    def __len__(self):
        return self.variables["N"].item()


class Group(VariableOwner, BrianObject):
    def _resolve(
        self, identifier, run_namespace, user_identifier=True, additional_variables=None
    ):
        """
        Resolve an identifier (i.e. variable, constant or function name) in the
        context of this group. This function will first lookup the name in the
        state variables, then look for a standard function or unit of that
        name and finally look in `Group.namespace` and in `run_namespace`. If
        the latter is not given, it will try to find the variable in the local
        namespace where the original function call took place. See
        :ref:`external-variables`.

        Parameters
        ----------
        identifiers : str
            The name to look up.
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        user_identifier : bool, optional
            Whether this is an identifier that was used by the user (and not
            something automatically generated that the user might not even
            know about). Will be used to determine whether to display a
            warning in the case of namespace clashes. Defaults to ``True``.
        additional_variables : dict-like, optional
            An additional mapping of names to `Variable` objects that will be
            checked before `Group.variables`.

        Returns
        -------
        obj : `Variable` or `Function`
            Returns a `Variable` object describing the variable or a `Function`
            object for a function. External variables are represented as
            `Constant` objects

        Raises
        ------
        KeyError
            If the `identifier` could not be resolved
        """
        resolved_internal = None

        if identifier in (additional_variables or {}):
            resolved_internal = additional_variables[identifier]
        elif identifier in getattr(self, "variables", {}):
            resolved_internal = self.variables[identifier]

        if resolved_internal is not None:
            if not user_identifier:
                return resolved_internal  # no need to go further
            # We already found the identifier, but we try to resolve it in the
            # external namespace nevertheless, to report a warning if it is
            # present there as well.
            self._resolve_external(
                identifier,
                run_namespace=run_namespace,
                internal_variable=resolved_internal,
            )
            return resolved_internal

        # We did not find the name internally, try to resolve it in the external
        # namespace
        return self._resolve_external(identifier, run_namespace=run_namespace)

    def resolve_all(
        self,
        identifiers,
        run_namespace,
        user_identifiers=None,
        additional_variables=None,
    ):
        """
        Resolve a list of identifiers. Calls `Group._resolve` for each
        identifier.

        Parameters
        ----------
        identifiers : iterable of str
            The names to look up.
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        user_identifiers : iterable of str, optional
            The names in ``identifiers`` that were provided by the user (i.e.
            are part of user-specified equations, abstract code, etc.). Will
            be used to determine when to issue namespace conflict warnings. If
            not specified, will be assumed to be identical to ``identifiers``.
        additional_variables : dict-like, optional
            An additional mapping of names to `Variable` objects that will be
            checked before `Group.variables`.

        Returns
        -------
        variables : dict of `Variable` or `Function`
            A mapping from name to `Variable`/`Function` object for each of the
            names given in `identifiers`

        Raises
        ------
        KeyError
            If one of the names in `identifier` cannot be resolved
        """
        if user_identifiers is None:
            user_identifiers = identifiers
        assert isinstance(run_namespace, Mapping)
        resolved = {}
        for identifier in identifiers:
            resolved[identifier] = self._resolve(
                identifier,
                user_identifier=identifier in user_identifiers,
                additional_variables=additional_variables,
                run_namespace=run_namespace,
            )
        return resolved

    def _resolve_external(
        self, identifier, run_namespace, user_identifier=True, internal_variable=None
    ):
        """
        Resolve an external identifier in the context of a `Group`. If the `Group`
        declares an explicit namespace, this namespace is used in addition to the
        standard namespace for units and functions. Additionally, the namespace in
        the `run_namespace` argument (i.e. the namespace provided to `Network.run`)
        is used.

        Parameters
        ----------
        identifier : str
            The name to resolve.
        group : `Group`
            The group that potentially defines an explicit namespace for looking up
            external names.
        run_namespace : dict
            A namespace (mapping from strings to objects), as provided as an
            argument to the `Network.run` function or returned by
            `get_local_namespace`.
        user_identifier : bool, optional
            Whether this is an identifier that was used by the user (and not
            something automatically generated that the user might not even
            know about). Will be used to determine whether to display a
            warning in the case of namespace clashes. Defaults to ``True``.
        internal_variable : `Variable`, optional
            The internal variable object that corresponds to this name (if any).
            This is used to give warnings if it also corresponds to a variable
            from an external namespace.
        """
        # We save tuples of (namespace description, referred object) to
        # give meaningful warnings in case of duplicate definitions
        matches = []

        namespaces = OrderedDict()
        # Default namespaces (units and functions)
        namespaces["constants"] = DEFAULT_CONSTANTS
        namespaces["units"] = DEFAULT_UNITS
        namespaces["functions"] = DEFAULT_FUNCTIONS
        if getattr(self, "namespace", None) is not None:
            namespaces["group-specific"] = self.namespace

        # explicit or implicit run namespace
        namespaces["run"] = run_namespace

        for description, namespace in namespaces.items():
            if identifier in namespace:
                match = namespace[identifier]
                if (
                    isinstance(
                        match,
                        (numbers.Number, np.ndarray, np.number, Function, Variable),
                    )
                ) or (
                    inspect.isfunction(match)
                    and hasattr(match, "_arg_units")
                    and hasattr(match, "_return_unit")
                ):
                    matches.append((description, match))

        if len(matches) == 0:
            # No match at all
            if internal_variable is not None:
                return None
            else:
                # Give a more detailed explanation for the lastupdate variable
                # that was removed with PR #1003
                if identifier == "lastupdate":
                    error_msg = (
                        'The identifier "lastupdate" could not be '
                        "resolved. Note that this variable is only "
                        "automatically defined for models with "
                        "event-driven synapses. You can define it "
                        'manually by adding "lastupdate : second" to '
                        'the equations and setting "lastupdate = t" '
                        "at the end of your on_pre and/or on_post "
                        "statements."
                    )
                else:
                    error_msg = f'The identifier "{identifier}" could not be resolved.'
                raise KeyError(error_msg)

        elif len(matches) > 1:
            # Possibly, all matches refer to the same object
            first_obj = matches[0][1]
            found_mismatch = False
            for m in matches:
                if _same_value(m[1], first_obj):
                    continue
                if _same_function(m[1], first_obj):
                    continue
                try:
                    proxy = weakref.proxy(first_obj)
                    if m[1] is proxy:
                        continue
                except TypeError:
                    pass

                # Found a mismatch
                found_mismatch = True
                break

            if found_mismatch and user_identifier and internal_variable is None:
                _conflict_warning(
                    'The name "%s" refers to different objects '
                    "in different namespaces used for resolving "
                    'names in the context of group "%s". '
                    "Will use the object from the %s namespace "
                    "with the value %s,"
                    % (
                        identifier,
                        getattr(self, "name", "<unknown>"),
                        matches[0][0],
                        _display_value(first_obj),
                    ),
                    matches[1:],
                )

        if internal_variable is not None and user_identifier:
            # Filter out matches that are identical (a typical case being an
            # externally defined "N" with the the number of neurons and a later
            # use of "N" in an expression (which refers to the internal variable
            # storing the number of neurons in the group)
            if isinstance(internal_variable, Constant):
                filtered_matches = []
                for match in matches:
                    if not _same_value(match[1], internal_variable):
                        filtered_matches.append(match)
            else:
                filtered_matches = matches
            if len(filtered_matches) == 0:
                pass  # Nothing to warn about
            else:
                warning_message = (
                    f"'{identifier}' is an internal variable of group "
                    f"'{self.name}', but also exists in the "
                )
                if len(matches) == 1:
                    namespace = filtered_matches[0][0]
                    value = _display_value(filtered_matches[0][1])
                    warning_message += f"{namespace} namespace with the value {value}. "
                else:
                    namespaces_list = " ,".join(match[0] for match in filtered_matches)
                    warning_message += f"following namespaces: {namespaces_list}. "
                warning_message += "The internal variable will be used."
                logger.warn(
                    warning_message, "Group.resolve.resolution_conflict", once=True
                )

        if internal_variable is not None:
            return None  # We were only interested in the warnings above

        # use the first match (according to resolution order)
        resolved = matches[0][1]

        # Replace pure Python functions by a Functions object
        if callable(resolved) and not isinstance(resolved, Function):
            resolved = Function(
                resolved,
                arg_units=getattr(resolved, "_arg_units", None),
                arg_names=getattr(resolved, "_arg_names", None),
                return_unit=getattr(resolved, "_return_unit", None),
                stateless=getattr(resolved, "stateless", False),
            )

        if not isinstance(resolved, (Function, Variable)):
            # Wrap the value in a Constant object
            dimensions = getattr(resolved, "dim", DIMENSIONLESS)
            value = np.asarray(resolved)
            if value.shape != ():
                raise KeyError(
                    f"Variable {identifier} was found in the namespace, but is not a"
                    " scalar value"
                )
            resolved = Constant(identifier, dimensions=dimensions, value=value)

        return resolved

    def runner(self, *args, **kwds):
        raise AttributeError("The 'runner' method has been renamed to 'run_regularly'.")

    def custom_operation(self, *args, **kwds):
        raise AttributeError(
            "The 'custom_operation' method has been renamed to 'run_regularly'."
        )

    def run_regularly(
        self,
        code,
        dt=None,
        clock=None,
        when="start",
        order=0,
        name=None,
        codeobj_class=None,
    ):
        """
        Run abstract code in the group's namespace. The created `CodeRunner`
        object will be automatically added to the group, it therefore does not
        need to be added to the network manually. However, a reference to the
        object will be returned, which can be used to later remove it from the
        group or to set it to inactive.

        Parameters
        ----------
        code : str
            The abstract code to run.
        dt : `Quantity`, optional
            The time step to use for this custom operation. Cannot be combined
            with the `clock` argument.
        clock : `Clock`, optional
            The update clock to use for this operation. If neither a clock nor
            the `dt` argument is specified, defaults to the clock of the group.
        when : str, optional
            When to run within a time step, defaults to the ``'start'`` slot.
            See :ref:`scheduling` for possible values.
        name : str, optional
            A unique name, if non is given the name of the group appended with
            'run_regularly', 'run_regularly_1', etc. will be used. If a
            name is given explicitly, it will be used as given (i.e. the group
            name will not be prepended automatically).
        codeobj_class : class, optional
            The `CodeObject` class to run code with. If not specified, defaults
            to the `group`'s ``codeobj_class`` attribute.

        Returns
        -------
        obj : `CodeRunner`
            A reference to the object that will be run.
        """
        if name is None:
            names = [o.name for o in self.contained_objects]
            name = find_name(f"{self.name}_run_regularly*", names)

        if dt is None and clock is None:
            clock = self._clock

        # Subgroups are normally not included in their parent's
        # contained_objects list, since there's no need to include them in the
        # network (they don't perform any computation on their own). However,
        # if a subgroup declares a `run_regularly` operation, then we want to
        # include this operation automatically, i.e. with the parent group
        # (adding just the run_regularly operation to the parent group's
        # contained objects would no be enough, since the CodeObject needs a
        # reference to the group providing the context for the operation, i.e.
        # the subgroup instead of the parent group. See github issue #922
        source_group = getattr(self, "source", None)
        if source_group is not None:
            if self not in source_group.contained_objects:
                source_group.contained_objects.append(self)

        runner = CodeRunner(
            self,
            "stateupdate",
            code=code,
            name=name,
            dt=dt,
            clock=clock,
            when=when,
            order=order,
            codeobj_class=codeobj_class,
        )
        self.contained_objects.append(runner)
        return runner

    def _check_for_invalid_states(self):
        """
        Checks if any state variables updated by differential equations have
        invalid values, and logs a warning if so.
        """
        equations = getattr(self, "equations", None)
        if not isinstance(equations, Equations):
            return
        for varname in equations.diff_eq_names:
            self._check_for_invalid_values(
                varname, self.state(varname, use_units=False)
            )

    def _check_for_invalid_values(self, k, v):
        """
        Checks if variable named k value v has invalid values, and logs a
        warning if so.
        """
        v = np.asarray(v)
        if np.isnan(v).any() or (np.abs(v) > 1e50).any():
            logger.warn(
                f"{self.name}'s variable '{k}' has NaN, very large values, "
                "or encountered an error in numerical integration. "
                "This is usually a sign that an unstable or invalid "
                "integration method was "
                "chosen.",
                name_suffix="invalid_values",
                once=True,
            )


class CodeRunner(BrianObject):
    """
    A "code runner" that runs a `CodeObject` every timestep and keeps a
    reference to the `Group`. Used in `NeuronGroup` for `Thresholder`,
    `Resetter` and `StateUpdater`.

    On creation, we try to run the before_run method with an empty additional
    namespace (see `Network.before_run`). If the namespace is already complete
    this might catch unit mismatches.

    Parameters
    ----------
    group : `Group`
        The group to which this object belongs.
    template : `Template`
        The template that should be used for code generation
    code : str, optional
        The abstract code that should be executed every time step. The
        `update_abstract_code` method might generate this code dynamically
        before every run instead.
    dt : `Quantity`, optional
        The time step to be used for the simulation. Cannot be combined with
        the `clock` argument.
    user_code : str, optional
        The abstract code as specified by the user, i.e. without any additions
        of internal code that the user not necessarily knows about. This will
        be used for warnings and error messages.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    when : str, optional
        In which scheduling slot to execute the operation during a time step.
        Defaults to ``'start'``. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of this operation for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    name : str, optional
        The name for this object.
    check_units : bool, optional
        Whether the units should be checked for consistency before a run. Is
        activated (``True``) by default but should be switched off for state
        updaters (units are already checked for the equations and the generated
        abstract code might have already replaced variables with their unit-less
        values)
    template_kwds : dict, optional
        A dictionary of additional information that is passed to the template.
    needed_variables: list of str, optional
        A list of variables that are neither present in the abstract code, nor
        in the ``USES_VARIABLES`` statement in the template. This is only
        rarely necessary, an example being a `StateMonitor` where the
        names of the variables are neither known to the template nor included
        in the abstract code statements.
    override_conditional_write: list of str, optional
        A list of variable names which are used as conditions (e.g. for
        refractoriness) which should be ignored.
    codeobj_class : class, optional
        The `CodeObject` class to run code with. If not specified, defaults to
        the `group`'s ``codeobj_class`` attribute.
    generate_empty_code : bool, optional
        Whether to generate a `CodeObject` if there is no abstract code to
        execute. Defaults to ``True`` but should be switched off e.g. for a
        `StateUpdater` when there is nothing to do.
    """

    add_to_magic_network = True
    invalidates_magic_network = True

    def __init__(
        self,
        group,
        template,
        code="",
        user_code=None,
        dt=None,
        clock=None,
        when="start",
        order=0,
        name="coderunner*",
        check_units=True,
        template_kwds=None,
        needed_variables=None,
        override_conditional_write=None,
        codeobj_class=None,
        generate_empty_code=True,
    ):
        BrianObject.__init__(
            self, clock=clock, dt=dt, when=when, order=order, name=name
        )
        self.group = weakproxy_with_fallback(group)
        self.template = template
        self.user_code = user_code
        self.abstract_code = code
        self.check_units = check_units
        if needed_variables is None:
            needed_variables = []
        self.needed_variables = needed_variables
        self.template_kwds = template_kwds
        self.override_conditional_write = override_conditional_write
        if codeobj_class is None:
            codeobj_class = group.codeobj_class
        self.codeobj_class = codeobj_class
        self.generate_empty_code = generate_empty_code
        self.codeobj = None

    def __getstate__(self):
        state = self.__dict__.copy()
        state["group"] = state["group"].__repr__.__self__
        return state

    def __setstate__(self, state):
        state["group"] = weakproxy_with_fallback(state["group"])
        self.__dict__ = state

    def update_abstract_code(self, run_namespace):
        """
        Update the abstract code for the code object. Will be called in
        `before_run` and should update the `CodeRunner.abstract_code`
        attribute.

        Does nothing by default.
        """
        pass

    def create_default_code_object(self, run_namespace):
        self.update_abstract_code(run_namespace=run_namespace)
        # If the CodeRunner has variables, add them
        if hasattr(self, "variables"):
            additional_variables = self.variables
        else:
            additional_variables = None

        if not self.generate_empty_code and len(self.abstract_code) == 0:
            self.codeobj = None
        else:
            self.codeobj = create_runner_codeobj(
                group=self.group,
                code=self.abstract_code,
                user_code=self.user_code,
                template_name=self.template,
                name=f"{self.name}_codeobject*",
                check_units=self.check_units,
                additional_variables=additional_variables,
                needed_variables=self.needed_variables,
                run_namespace=run_namespace,
                template_kwds=self.template_kwds,
                override_conditional_write=self.override_conditional_write,
                codeobj_class=self.codeobj_class,
            )
        return self.codeobj

    def create_code_objects(self, run_namespace):
        # By default, we only have one code object for each CodeRunner.
        # Overwrite this function to use more than one.
        code_object = self.create_default_code_object(run_namespace)
        if code_object:
            self.code_objects[:] = [weakref.proxy(code_object)]
        else:
            self.code_objects[:] = []

    def before_run(self, run_namespace):
        self.create_code_objects(run_namespace)
        super().before_run(run_namespace)



"""
A pseudo MSO neuron, with two dendrites (fake geometry).
There are synaptic inputs.

Second method.
"""
from brian2 import *

# Morphology
morpho = Soma(30*um)
morpho.L = Cylinder(diameter=1*um, length=100*um, n=50)
morpho.R = Cylinder(diameter=1*um, length=100*um, n=50)

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
Es = 0*mV
taus = 1*ms
eqs='''
Im = gL*(EL-v) : amp/meter**2
Is = gs*(Es-v) : amp (point current)
dgs/dt = -gs/taus : siemens
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs,
                       Cm=1*uF/cm**2, Ri=100*ohm*cm, method='exponential_euler')
neuron.v = EL

# Regular inputs
stimulation = NeuronGroup(2, 'dx/dt = 300*Hz : 1', threshold='x>1', reset='x=0',
                          method='euler')
stimulation.x = [0, 0.5] # Asynchronous

# Synapses
w = 20*nS
S = Synapses(stimulation, neuron, on_pre='gs += w')
S.connect(i=0, j=morpho.L[99.9*um])
S.connect(i=1, j=morpho.R[99.9*um])

# Monitors
mon_soma = StateMonitor(neuron, 'v', record=[0])
mon_L = StateMonitor(neuron.L, 'v', record=True)
mon_R = StateMonitor(neuron, 'v', record=morpho.R[99.9*um])

run(50*ms, report='text')

subplot(211)
plot(mon_L.t/ms, mon_soma[0].v/mV, 'k')
plot(mon_L.t/ms, mon_L[morpho.L[99.9*um]].v/mV, 'r')
plot(mon_L.t/ms, mon_R[morpho.R[99.9*um]].v/mV, 'b')
ylabel('v (mV)')
subplot(212)
for i in [0, 5, 10, 15, 20, 25, 30, 35, 40, 45]:
    plot(mon_L.t/ms, mon_L.v[i, :]/mV)
xlabel('Time (ms)')
ylabel('v (mV)')
show()



#!/usr/bin/env python3
"""
Reproduces Figure 4B from Stimberg et al. (2019).

In contrast to the original example, this version uses the standalone mode of Brian 2, using the new ``run_args``
feature introduced in version 2.6.

Marcel Stimberg, Romain Brette, Dan FM Goodman (2019)
Brian 2, an intuitive and efficient neural simulator eLife 8:e47314

https://doi.org/10.7554/eLife.47314
"""

from brian2 import *

defaultclock.dt = 0.01*ms  # small time step for stiff equations
set_device('cpp_standalone', build_on_run=False)

# Our model of the neuron is based on the classical model of from Hodgkin and Huxley (1952). Note that this is not
# actually a model of a neuron, but rather of a (space-clamped) axon. However, to avoid confusion with spatially
# extended models, we simply use the term "neuron" here. In this model, the membrane potential is shifted, i.e. the
# resting potential is at 0mV:
El = 10.613*mV
ENa = 115*mV
EK = -12*mV
gl = 0.3*msiemens/cm**2
gK = 36*msiemens/cm**2
gNa_max = 100*msiemens/cm**2
gNa_min = 15*msiemens/cm**2
C = 1*uF/cm**2

eqs = """
dv/dt = (gl * (El-v) + gNa * m**3 * h * (ENa-v) + gK * n**4 * (EK-v)) / C : volt
gNa : siemens/meter**2 (constant)
dm/dt = alpham * (1-m) - betam * m : 1
dn/dt = alphan * (1-n) - betan * n : 1
dh/dt = alphah * (1-h) - betah * h : 1
alpham = (0.1/mV) * 10*mV / exprel((-v+25*mV) / (10*mV))/ms : Hz
betam = 4 * exp(-v/(18*mV))/ms : Hz
alphah = 0.07 * exp(-v/(20*mV))/ms : Hz
betah = 1/(exp((-v+30*mV) / (10*mV)) + 1)/ms : Hz
alphan = (0.01/mV) * 10*mV / exprel((-v+10*mV) / (10*mV))/ms : Hz
betan = 0.125*exp(-v/(80*mV))/ms : Hz
"""

# We simulate 100 neurons at the same time, each of them having a density of sodium channels between 15 and 100 mS/cm²:
neurons = NeuronGroup(100, eqs, method="rk4", threshold="v>50*mV", reset="")
neurons.gNa = "gNa_min + (gNa_max - gNa_min)*1.0*i/N"

# We initialize the state variables to their resting state values, note that the values for $m$, $n$, $h$ depend on the
# values of $\alpha_m$, $\beta_m$, etc. which themselves depend on $v$. The order of the assignments ($v$ is
# initialized before $m$, $n$, and $h$) therefore matters, something that is naturally expressed by stating initial
# values as sequential assignments to the state variables. In a declarative approach, this would be potentially
# ambiguous.

neurons.v = 0*mV
# Note that the initial values of the gating variables are fixed here, i.e. they are relative to the resting potential,
# not to the membrane potential we set to check for the threshold. If we wanted this to be the case, we'd have to call
# `device.apply_run_args()` here, so that the value we provide is applied before initializing m, n, and h.
neurons.m = "1/(1 + betam/alpham)"
neurons.n = "1/(1 + betan/alphan)"
neurons.h = "1/(1 + betah/alphah)"

S = SpikeMonitor(neurons)

# The algorithm we use here to find the voltage threshold is a simple bisection: we try to find the threshold voltage of
# a neuron by repeatedly testing values and increasing or decreasing these values depending on whether we observe a
# spike or not. By continously halving the size of the correction, we quickly converge to a precise estimate.
# 
# We start with the same initial estimate for all segments, 25mV above the resting potential, and the same value for
# the size of the "correction step":

v0 = 25*mV*np.ones(len(neurons))
step = 25*mV

# For later visualization of how the estimates converged towards their final values, we also store the intermediate values of the estimates:
estimates = np.full((11, len(neurons)), np.nan)*mV
estimates[0, :] = v0
# Run the simulation for 20ms
run(20*ms)
device.build(run=False, directory=None)

for i in range(10):
    print(".", end="")
    # Set the membrane potential to our threshold estimate
    device.run(run_args={neurons.v: v0})
    # Decrease the estimates for neurons that spiked
    v0[S.count > 0] -= step
    # Increase the estimate for neurons that did not spike
    v0[S.count == 0] += step
    # Reduce step size and store current estimate
    step /= 2.0
    estimates[i + 1, :] = v0
print()

# After the 10 iteration steps, we plot the results:
plt.rcParams.update({'axes.spines.top': False,
                     'axes.spines.right': False})
fig, (ax1, ax2) = plt.subplots(1, 2, sharey=True, figsize=(8, 3), layout="constrained")

colors = ["#1f77b4", "#ff7f03", "#2ca02c"]
examples = [10, 50, 90]
for example, color in zip(examples, colors):
    ax1.plot(np.arange(11), estimates[:, example] / mV,
             "o-", mec='white', lw=2, color=color, clip_on=False, ms=10, zorder=100,
             label=f"gNA = {(neurons.gNa[example]/(mS/cm**2)):.1f}mS/cm$^2$")

ax2.plot(neurons.gNa/(mS/cm**2), v0/mV, color="gray", lw=2)

for idx, (example, color) in enumerate(zip(examples, colors)):
    ax2.plot([neurons.gNa[example]/(mS/cm**2)], [estimates[-1, example]/mV], "o", color=color, mec="white", ms=10,
             label=f"gNA = {(neurons.gNa[example]/(mS/cm**2)):.1f}mS/cm$^2$")

ax1.set(title="iteration", ylim=(0, 45), ylabel="threshold estimate (mV)")
ax2.set(title="gNA (mS/cm²)")

plt.show()



# coding=utf-8
"""
Modeling neuron-glia interactions with the Brian 2 simulator
Marcel Stimberg, Dan F. M. Goodman, Romain Brette, Maurizio De Pittà
bioRxiv 198366; doi: https://doi.org/10.1101/198366

Figure 6: Recurrent neuron-glial network.

Randomly connected COBA network (see Brunel, 2000) with excitatory synapses modulated
by release-increasing gliotransmission from a randomly connected network of astrocytes.
"""

from brian2 import *

import plot_utils as pu

set_device('cpp_standalone', directory=None)  # Use fast "C++ standalone mode"
seed(28371)  # to get identical figures for repeated runs

################################################################################
# Model parameters
################################################################################
### General parameters
N_e = 3200                   # Number of excitatory neurons
N_i = 800                    # Number of inhibitory neurons
N_a = 3200                   # Number of astrocytes

## Some metrics parameters needed to establish proper connections
size = 3.75*mmeter           # Length and width of the square lattice
distance = 50*umeter         # Distance between neurons

### Neuron parameters
E_l = -60*mV                 # Leak reversal potential
g_l = 9.99*nS                # Leak conductance
E_e = 0*mV                   # Excitatory synaptic reversal potential
E_i = -80*mV                 # Inhibitory synaptic reversal potential
C_m = 198*pF                 # Membrane capacitance
tau_e = 5*ms                 # Excitatory synaptic time constant
tau_i = 10*ms                # Inhibitory synaptic time constant
tau_r = 5*ms                 # Refractory period
I_ex = 100*pA                # External current
V_th = -50*mV                # Firing threshold
V_r = E_l                    # Reset potential

### Synapse parameters
rho_c = 0.005                # Synaptic vesicle-to-extracellular space volume ratio
Y_T = 500.*mmolar            # Total vesicular neurotransmitter concentration
Omega_c = 40/second          # Neurotransmitter clearance rate
U_0__star = 0.6              # Resting synaptic release probability
Omega_f = 3.33/second        # Synaptic facilitation rate
Omega_d = 2.0/second         # Synaptic depression rate
w_e = 0.05*nS                # Excitatory synaptic conductance
w_i = 1.0*nS                 # Inhibitory synaptic conductance
# --- Presynaptic receptors
O_G = 1.5/umolar/second      # Agonist binding (activating) rate
Omega_G = 0.5/(60*second)    # Agonist release (deactivating) rate

### Astrocyte parameters
# ---  Calcium fluxes
O_P = 0.9*umolar/second      # Maximal Ca^2+ uptake rate by SERCAs
K_P = 0.05*umolar            # Ca2+ affinity of SERCAs
C_T = 2*umolar               # Total cell free Ca^2+ content
rho_A = 0.18                 # ER-to-cytoplasm volume ratio
Omega_C = 6/second           # Maximal rate of Ca^2+ release by IP_3Rs
Omega_L = 0.1/second         # Maximal rate of Ca^2+ leak from the ER
# --- IP_3R kinectics
d_1 = 0.13*umolar            # IP_3 binding affinity
d_2 = 1.05*umolar            # Ca^2+ inactivation dissociation constant
O_2 = 0.2/umolar/second      # IP_3R binding rate for Ca^2+ inhibition
d_3 = 0.9434*umolar          # IP_3 dissociation constant
d_5 = 0.08*umolar            # Ca^2+ activation dissociation constant
# --- IP_3 production
# --- Agonist-dependent IP_3 production
O_beta = 0.5*umolar/second   # Maximal rate of IP_3 production by PLCbeta
O_N = 0.3/umolar/second      # Agonist binding rate
Omega_N = 0.5/second         # Maximal inactivation rate
K_KC = 0.5*umolar            # Ca^2+ affinity of PKC
zeta = 10                    # Maximal reduction of receptor affinity by PKC
# --- Endogenous IP3 production
O_delta = 1.2*umolar/second  # Maximal rate of IP_3 production by PLCdelta
kappa_delta = 1.5*umolar     # Inhibition constant of PLC_delta by IP_3
K_delta = 0.1*umolar         # Ca^2+ affinity of PLCdelta
# --- IP_3 degradation
Omega_5P = 0.05/second       # Maximal rate of IP_3 degradation by IP-5P
K_D = 0.7*umolar             # Ca^2+ affinity of IP3-3K
K_3K = 1.0*umolar            # IP_3 affinity of IP_3-3K
O_3K = 4.5*umolar/second     # Maximal rate of IP_3 degradation by IP_3-3K
# --- IP_3 diffusion
F = 0.09*umolar/second       # GJC IP_3 permeability
I_Theta = 0.3*umolar         # Threshold gradient for IP_3 diffusion
omega_I = 0.05*umolar        # Scaling factor of diffusion
# --- Gliotransmitter release and time course
C_Theta = 0.5*umolar         # Ca^2+ threshold for exocytosis
Omega_A = 0.6/second         # Gliotransmitter recycling rate
U_A = 0.6                    # Gliotransmitter release probability
G_T = 200*mmolar             # Total vesicular gliotransmitter concentration
rho_e = 6.5e-4               # Astrocytic vesicle-to-extracellular volume ratio
Omega_e = 60/second          # Gliotransmitter clearance rate
alpha = 0.0                  # Gliotransmission nature

################################################################################
# Define HF stimulus
################################################################################
stimulus = TimedArray([1.0, 1.2, 1.0, 1.0], dt=2*second)

################################################################################
# Simulation time (based on the stimulus)
################################################################################
duration = 8*second          # Total simulation time

################################################################################
# Model definition
################################################################################
### Neurons
neuron_eqs = '''
dv/dt = (g_l*(E_l-v) + g_e*(E_e-v) + g_i*(E_i-v) + I_ex*stimulus(t))/C_m : volt (unless refractory)
dg_e/dt = -g_e/tau_e : siemens  # post-synaptic excitatory conductance
dg_i/dt = -g_i/tau_i : siemens  # post-synaptic inhibitory conductance
# Neuron position in space
x : meter (constant)
y : meter (constant)
'''
neurons = NeuronGroup(N_e + N_i, model=neuron_eqs,
                      threshold='v>V_th', reset='v=V_r',
                      refractory='tau_r', method='euler')
exc_neurons = neurons[:N_e]
inh_neurons = neurons[N_e:]
# Arrange excitatory neurons in a grid
N_rows = int(sqrt(N_e))
N_cols = N_e//N_rows
grid_dist = (size / N_cols)
exc_neurons.x = '(i // N_rows)*grid_dist - N_rows/2.0*grid_dist'
exc_neurons.y = '(i % N_rows)*grid_dist - N_cols/2.0*grid_dist'
# Random initial membrane potential values and conductances
neurons.v = 'E_l + rand()*(V_th-E_l)'
neurons.g_e = 'rand()*w_e'
neurons.g_i = 'rand()*w_i'

### Synapses
synapses_eqs = '''
# Neurotransmitter
dY_S/dt = -Omega_c * Y_S                                    : mmolar (clock-driven)
# Fraction of activated presynaptic receptors
dGamma_S/dt = O_G * G_A * (1 - Gamma_S) - Omega_G * Gamma_S : 1 (clock-driven)
# Usage of releasable neurotransmitter per single action potential:
du_S/dt = -Omega_f * u_S                                    : 1 (event-driven)
# Fraction of synaptic neurotransmitter resources available for release:
dx_S/dt = Omega_d *(1 - x_S)                                : 1 (event-driven)
U_0                                                         : 1
# released synaptic neurotransmitter resources:
r_S                                                         : 1
# gliotransmitter concentration in the extracellular space:
G_A                                                         : mmolar
# which astrocyte covers this synapse ?
astrocyte_index : integer (constant)
'''
synapses_action = '''
U_0 = (1 - Gamma_S) * U_0__star + alpha * Gamma_S
u_S += U_0 * (1 - u_S)
r_S = u_S * x_S
x_S -= r_S
Y_S += rho_c * Y_T * r_S
'''
exc_syn = Synapses(exc_neurons, neurons, model=synapses_eqs,
                   on_pre=synapses_action+'g_e_post += w_e*r_S',
                   method='exact')
exc_syn.connect(True, p=0.05)
exc_syn.x_S = 1.0
inh_syn = Synapses(inh_neurons, neurons, model=synapses_eqs,
                   on_pre=synapses_action+'g_i_post += w_i*r_S',
                   method='exact')
inh_syn.connect(True, p=0.2)
inh_syn.x_S = 1.0
# Connect excitatory synapses to an astrocyte depending on the position of the
# post-synaptic neuron
N_rows_a = int(sqrt(N_a))
N_cols_a = N_a/N_rows_a
grid_dist = size / N_rows_a
exc_syn.astrocyte_index = ('int(x_post/grid_dist) + '
                           'N_cols_a*int(y_post/grid_dist)')
### Astrocytes
# The astrocyte emits gliotransmitter when its Ca^2+ concentration crosses
# a threshold
astro_eqs = '''
# Fraction of activated astrocyte receptors:
dGamma_A/dt = O_N * Y_S * (1 - clip(Gamma_A,0,1)) -
              Omega_N*(1 + zeta * C/(C + K_KC)) * clip(Gamma_A,0,1) : 1
# Intracellular IP_3
dI/dt = J_beta + J_delta - J_3K - J_5P + J_coupling              : mmolar
J_beta = O_beta * Gamma_A                                        : mmolar/second
J_delta = O_delta/(1 + I/kappa_delta) * C**2/(C**2 + K_delta**2) : mmolar/second
J_3K = O_3K * C**4/(C**4 + K_D**4) * I/(I + K_3K)                : mmolar/second
J_5P = Omega_5P*I                                                : mmolar/second
# Diffusion between astrocytes:
J_coupling                                                       : mmolar/second

# Ca^2+-induced Ca^2+ release:
dC/dt = J_r + J_l - J_p                                   : mmolar
dh/dt = (h_inf - h)/tau_h                                 : 1
J_r = (Omega_C * m_inf**3 * h**3) * (C_T - (1 + rho_A)*C) : mmolar/second
J_l = Omega_L * (C_T - (1 + rho_A)*C)                     : mmolar/second
J_p = O_P * C**2/(C**2 + K_P**2)                          : mmolar/second
m_inf = I/(I + d_1) * C/(C + d_5)                         : 1
h_inf = Q_2/(Q_2 + C)                                     : 1
tau_h = 1/(O_2 * (Q_2 + C))                               : second
Q_2 = d_2 * (I + d_1)/(I + d_3)                           : mmolar

# Fraction of gliotransmitter resources available for release:
dx_A/dt = Omega_A * (1 - x_A) : 1
# gliotransmitter concentration in the extracellular space:
dG_A/dt = -Omega_e*G_A        : mmolar
# Neurotransmitter concentration in the extracellular space:
Y_S                           : mmolar
# The astrocyte position in space
x : meter (constant)
y : meter (constant)
'''
glio_release = '''
G_A += rho_e * G_T * U_A * x_A
x_A -= U_A *  x_A
'''
astrocytes = NeuronGroup(N_a, astro_eqs,
                         # The following formulation makes sure that a "spike" is
                         # only triggered at the first threshold crossing
                         threshold='C>C_Theta',
                         refractory='C>C_Theta',
                         # The gliotransmitter release happens when the threshold
                         # is crossed, in Brian terms it can therefore be
                         # considered a "reset"
                         reset=glio_release,
                         method='rk4',
                         dt=1e-2*second)
# Arrange astrocytes in a grid
astrocytes.x = '(i // N_rows_a)*grid_dist - N_rows_a/2.0*grid_dist'
astrocytes.y = '(i % N_rows_a)*grid_dist - N_cols_a/2.0*grid_dist'
# Add random initialization
astrocytes.C = 0.01*umolar
astrocytes.h = 0.9
astrocytes.I = 0.01*umolar
astrocytes.x_A = 1.0

ecs_astro_to_syn = Synapses(astrocytes, exc_syn,
                            'G_A_post = G_A_pre : mmolar (summed)')
ecs_astro_to_syn.connect('i == astrocyte_index_post')
ecs_syn_to_astro = Synapses(exc_syn, astrocytes,
                            'Y_S_post = Y_S_pre/N_incoming : mmolar (summed)')
ecs_syn_to_astro.connect('astrocyte_index_pre == j')
# Diffusion between astrocytes
astro_to_astro_eqs = '''
delta_I = I_post - I_pre            : mmolar
J_coupling_post = -(1 + tanh((abs(delta_I) - I_Theta)/omega_I))*
                  sign(delta_I)*F/2 : mmolar/second (summed)
'''
astro_to_astro = Synapses(astrocytes, astrocytes,
                          model=astro_to_astro_eqs)
# Connect to all astrocytes less than 75um away
# (about 4 connections per astrocyte)
astro_to_astro.connect('i != j and '
                       'sqrt((x_pre-x_post)**2 +'
                       '     (y_pre-y_post)**2) < 75*um')

################################################################################
# Monitors
################################################################################
# Note that we could use a single monitor for all neurons instead, but this
# way plotting is a bit easier in the end
exc_mon = SpikeMonitor(exc_neurons)
inh_mon = SpikeMonitor(inh_neurons)
ast_mon = SpikeMonitor(astrocytes)

################################################################################
# Simulation run
################################################################################
run(duration, report='text')

################################################################################
# Plot of Spiking activity
################################################################################
plt.style.use('figures.mplstyle')

fig, ax = plt.subplots(nrows=3, ncols=1, sharex=True, figsize=(6.26894, 6.26894*0.8),
                       gridspec_kw={'height_ratios': [1, 6, 2],
                                    'left': 0.12, 'top': 0.97})
time_range = np.linspace(0, duration/second, int(duration/second*100))*second
ax[0].plot(time_range, I_ex*stimulus(time_range)/pA, 'k')
ax[0].set(xlim=(0, duration/second), ylim=(98, 122),
          yticks=[100, 120], ylabel='$I_{ex}$ (pA)')
pu.adjust_spines(ax[0], ['left'])

## We only plot a fraction of the spikes
fraction = 4
ax[1].plot(exc_mon.t[exc_mon.i <= N_e//fraction]/second,
           exc_mon.i[exc_mon.i <= N_e//fraction], '|', color='C0')
ax[1].plot(inh_mon.t[inh_mon.i <= N_i//fraction]/second,
           inh_mon.i[inh_mon.i <= N_i//fraction]+N_e//fraction, '|', color='C1')
ax[1].plot(ast_mon.t[ast_mon.i <= N_a//fraction]/second,
           ast_mon.i[ast_mon.i <= N_a//fraction]+(N_e+N_i)//fraction,
           '|', color='C2')
ax[1].set(xlim=(0, duration/second), ylim=[0, (N_e+N_i+N_a)//fraction],
          yticks=np.arange(0, (N_e+N_i+N_a)//fraction+1, 250),
          ylabel='cell index')
pu.adjust_spines(ax[1], ['left'])

# Generate frequencies
bin_size = 1*ms
spk_count, bin_edges = np.histogram(np.r_[exc_mon.t/second, inh_mon.t/second],
                                    int(duration/bin_size))
rate = 1.0*spk_count/(N_e + N_i)/bin_size/Hz
rate[rate<0.001] = 0.001 # Fix 0 lower bound for log scale
ax[2].semilogy(bin_edges[:-1], rate, '-', color='k')
pu.adjust_spines(ax[2], ['left', 'bottom'])
ax[2].set(xlim=(0, duration/second), ylim=(0.1, 150),
          xticks=np.arange(0, 9), yticks=[0.1, 1, 10, 100],
          xlabel='time (s)', ylabel='rate (Hz)')
ax[2].get_yaxis().set_major_formatter(ScalarFormatter())

pu.adjust_ylabels(ax, x_offset=-0.11)

plt.show()



"""
This is a Brian script implementing a benchmark described
in the following review paper:

Simulation of networks of spiking neurons: A review of tools and strategies
(2007). Brette, Rudolph, Carnevale, Hines, Beeman, Bower, Diesmann, Goodman,
Harris, Zirpe, Natschlager, Pecevski, Ermentrout, Djurfeldt, Lansner, Rochel,
Vibert, Alvarez, Muller, Davison, El Boustani and Destexhe.
Journal of Computational Neuroscience 23(3):349-98

Benchmark 2: random network of integrate-and-fire neurons with exponential
synaptic currents.

Clock-driven implementation with exact subthreshold integration
(but spike times are aligned to the grid).
"""
from brian2 import *

taum = 20*ms
taue = 5*ms
taui = 10*ms
Vt = -50*mV
Vr = -60*mV
El = -49*mV

eqs = '''
dv/dt  = (ge+gi-(v-El))/taum : volt (unless refractory)
dge/dt = -ge/taue : volt
dgi/dt = -gi/taui : volt
'''

P = NeuronGroup(4000, eqs, threshold='v>Vt', reset='v = Vr', refractory=5*ms,
                method='exact')
P.v = 'Vr + rand() * (Vt - Vr)'
P.ge = 0*mV
P.gi = 0*mV

we = (60*0.27/10)*mV # excitatory synaptic weight (voltage)
wi = (-20*4.5/10)*mV # inhibitory synaptic weight
Ce = Synapses(P, P, on_pre='ge += we')
Ci = Synapses(P, P, on_pre='gi += wi')
Ce.connect('i<3200', p=0.02)
Ci.connect('i>=3200', p=0.02)

s_mon = SpikeMonitor(P)

run(1 * second)

plot(s_mon.t/ms, s_mon.i, ',k')
xlabel('Time (ms)')
ylabel('Neuron index')
show()



import pickle

import matplotlib.pyplot as plt
import numpy as np

with open('synapse_creation_times_brian1.pickle', 'r') as f:
    brian1_results = pickle.load(f)
with open('synapse_creation_times_brian2.pickle', 'r') as f:
    brian2_results = pickle.load(f)

results = brian1_results
results.update(brian2_results)

conditions = [('Full', True),
              ('Full (no-self)', 'i != j'),
              ('One-to-one', 'i == j'),
              ('Simple neighbourhood', 'abs(i-j) < 5'),
              ('Gauss neighbourhood', 'exp(-(i - j)**2/5) > 0.005'),
              ('Random (50%)', 0.5),
              ('Random (10%)', 0.1),
              ('Random (1%)', 0.01),
              ('Random no-self (50%)', '(i != j) * 0.5'),
              ('Random no-self (10%)', '(i != j) * 0.1'),
              ('Random no-self (1%)', '(i != j) * 0.01')]
targets = ['PythonLanguage', 'CPPLanguage', 'Brian 1']
# nicer names for the labels
lang_translation = {'PythonLanguage': 'Brian 2 (Python)',
                    'CPPLanguage': 'Brian 2 (C++)',
                    'Brian 1': 'Brian 1'}

# Do some plots
for pattern, condition in conditions:
    plt.figure()
    for lang_name in targets:
        data = [(connections, time) for ((lang, connections, p), time) in results.items()
                if lang == lang_name and p == pattern]
        data.sort(key=lambda item: item[0])
        data = np.array(data).T
        plt.plot(data[0], data[1], 'o-', label=lang_translation[lang_name])
    plt.xscale('log')
    plt.yscale('log')
    plt.xlabel('Number of created connections')
    plt.ylabel('time (s)')
    plt.legend(loc='best', frameon=False)
    plt.title('%s: "%s"' % (pattern, condition))
    plt.savefig('plots/%s.png' % pattern.replace(' ', '_'))


#!/usr/bin/env python
"""
Slope-threshold relationship with noisy inputs, in the adaptive threshold model
-------------------------------------------------------------------------------
Fig. 5E,F from:

Platkiewicz J and R Brette (2011). Impact of Fast Sodium Channel Inactivation on Spike
Threshold Dynamics and Synaptic Integration. PLoS Comp Biol 7(5):
e1001129. doi:10.1371/journal.pcbi.1001129
"""
from scipy import optimize
from scipy.stats import linregress

from brian2 import *

N = 200  # 200 neurons to get more statistics, only one is shown
duration = 1*second
# --Biophysical parameters
ENa = 60*mV
EL = -70*mV
vT = -55*mV
Vi = -63*mV
tauh = 5*ms
tau = 5*ms
ka = 5*mV
ki = 6*mV
a = ka / ki
tauI = 5*ms
mu = 15*mV
sigma = 6*mV / sqrt(tauI / (tauI + tau))

# --Theoretical prediction for the slope-threshold relationship (approximation: a=1+epsilon)
thresh = lambda slope, a: Vi - slope * tauh * log(1 + (Vi - vT / a) / (slope * tauh))
# -----Exact calculation of the slope-threshold relationship
# (note that optimize.fsolve does not work with units, we therefore let th be a
# unitless quantity, i.e. the value in volt).
thresh_ex = lambda s: optimize.fsolve(lambda th: (a*s*tauh*exp((Vi-th*volt)/(s*tauh))-th*volt*(1-a)-a*(s*tauh+Vi)+vT)/volt,
                                    thresh(s, a))*volt

eqs = """
dv/dt=(EL-v+mu+sigma*I)/tau : volt
dtheta/dt=(vT+a*clip(v-Vi, 0*mV, inf*mV)-theta)/tauh : volt
dI/dt=-I/tauI+(2/tauI)**.5*xi : 1 # Ornstein-Uhlenbeck
"""
neurons = NeuronGroup(N, eqs, threshold="v>theta", reset='v=EL',
                      refractory=5*ms)
neurons.v = EL
neurons.theta = vT
neurons.I = 0
S = SpikeMonitor(neurons)
M = StateMonitor(neurons, 'v', record=True)
Mt = StateMonitor(neurons, 'theta', record=0)

run(duration, report='text')

# Linear regression gives depolarization slope before spikes
tx = M.t[(M.t > 0*second) & (M.t < 1.5 * tauh)]
slope, threshold = [], []

for (i, t) in zip(S.i, S.t):
    ind = (M.t < t) & (M.t > t - tauh)
    mx = M.v[i, ind]
    s, _, _, _, _ = linregress(tx[:len(mx)]/ms, mx/mV)
    slope.append(s)
    threshold.append(mx[-1])

# Figure
fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4))

ax1.plot(M.t/ms, M.v[0]/mV, 'k')
ax1.plot(Mt.t/ms, Mt.theta[0]/mV, 'r')
# Display spikes on the trace
spike_timesteps = np.round(S.t[S.i == 0]/defaultclock.dt).astype(int)
ax1.vlines(S.t[S.i == 0]/ms,
           M.v[0, spike_timesteps]/mV,
           0, color='r')
ax1.plot(S.t[S.i == 0]/ms, M.v[0, spike_timesteps]/mV, 'ro', ms=3)
ax1.set(xlabel='Time (ms)', ylabel='Voltage (mV)', xlim=(0, 500),
        ylim=(-75, -35))

ax2.plot(slope, Quantity(threshold)/mV, 'r.')
sx = linspace(0.5*mV/ms, 4*mV/ms, 100)
t = Quantity([thresh_ex(s) for s in sx])
ax2.plot(sx/(mV/ms), t/mV, 'k')
ax2.set(xlim=(0.5, 4), xlabel='Depolarization slope (mV/ms)',
        ylabel='Threshold (mV)')

fig.tight_layout()
plt.show()



#!/usr/bin/env python3
"""Hodgkin-Huxley neuron simulation with approximations for gating variable steady-states and time constants

Follows exercise 4, chapter 2 of Eugene M. Izhikevich: Dynamical Systems in Neuroscience

Sebastian Schmitt, 2021
"""

import argparse
from functools import reduce
import operator

import matplotlib.pyplot as plt
from cycler import cycler
import numpy as np

from brian2 import run
from brian2 import mS, cmeter, ms, mV, uA, uF
from brian2 import Equations, NeuronGroup, StateMonitor, TimedArray, defaultclock


def construct_gating_variable_inf_equation(gating_variable):
    """Construct the voltage-dependent steady-state gating variable equation.

    Approximated by Boltzmann function.

    gating_variable -- gating variable, typically one of "m", "n" and "h"
    """

    return Equations('xinf = 1/(1+exp((v_half-v)/k)) : 1',
                     xinf=f'{gating_variable}_inf',
                     v_half=f'v_{gating_variable}_half',
                     k=f'k_{gating_variable}')


def construct_gating_variable_tau_equation(gating_variable):
    """Construct the voltage-dependent gating variable time constant equation.

    Approximated by Gaussian function.

    gating_variable -- gating variable, typically one of "m", "n" and "h"
    """

    return Equations('tau = c_base + c_amp*exp(-(v_max - v)**2/sigma**2) : second',
                     tau=f'tau_{gating_variable}',
                     c_base=f'c_{gating_variable}_base',
                     c_amp=f'c_{gating_variable}_amp',
                     v_max=f'v_{gating_variable}_max',
                     sigma=f'sigma_{gating_variable}')


def construct_gating_variable_ode(gating_variable):
    """Construct the ordinary differential equation of the gating variable.

    gating_variable -- gating variable, typically one of "m", "n" and "h"
    """

    return Equations('dx/dt = (xinf - x)/tau : 1',
                     x=gating_variable,
                     xinf=f'{gating_variable}_inf',
                     tau=f'tau_{gating_variable}')


def construct_neuron_ode():
    """Construct the ordinary differential equation of the membrane."""

    # conductances
    g_K_eq = Equations('g_K = g_K_bar*n**4 : siemens/meter**2')
    g_Na_eq = Equations('g_Na = g_Na_bar*m**3*h : siemens/meter**2')

    # currents
    I_K_eq = Equations('I_K = g_K*(v - e_K) : ampere/meter**2')
    I_Na_eq = Equations('I_Na = g_Na*(v - e_Na) : ampere/meter**2')
    I_L_eq = Equations('I_L = g_L*(v - e_L) : ampere/meter**2')

    # external drive
    I_ext_eq = Equations('I_ext = I_stim(t) : ampere/meter**2')

    # membrane
    membrane_eq = Equations('dv/dt = (I_ext - I_K - I_Na - I_L)/C_mem : volt')

    return [g_K_eq, g_Na_eq, I_K_eq, I_Na_eq, I_L_eq, I_ext_eq, membrane_eq]


def plot_tau(ax, parameters):
    """Plot gating variable time constants as function of membrane potential.

    ax -- matplotlib axes to be plotted on
    parameters -- dictionary of parameters for gating variable time constant equations
    """

    tau_group = NeuronGroup(100,
                            Equations('v : volt') +
                            reduce(operator.add, [construct_gating_variable_tau_equation(
                                gv) for gv in ['m', 'n', 'h']]),
                            method='euler', namespace=parameters)

    min_v = -100
    max_v = 100
    tau_group.v = np.linspace(min_v, max_v, len(tau_group))*mV

    ax.plot(tau_group.v/mV, tau_group.tau_m/ms, label=r'$\tau_m$')
    ax.plot(tau_group.v/mV, tau_group.tau_n/ms, label=r'$\tau_n$')
    ax.plot(tau_group.v/mV, tau_group.tau_h/ms, label=r'$\tau_h$')

    ax.set_xlabel('$v$ (mV)')
    ax.set_ylabel(r'$\tau$ (ms)')
    ax.yaxis.set_label_position("right")
    ax.yaxis.tick_right()
    ax.legend()


def plot_inf(ax, parameters):
    """Plot gating variable steady-state values as function of membrane potential.

    ax -- matplotlib axes to be plotted on
    parameters -- dictionary of parameters for gating variable steady-state equations
    """

    inf_group = NeuronGroup(100,
                            Equations('v : volt') +
                            reduce(operator.add, [construct_gating_variable_inf_equation(
                                gv) for gv in ['m', 'n', 'h']]),
                            method='euler', namespace=parameters)
    inf_group.v = np.linspace(-100, 100, len(inf_group))*mV

    ax.plot(inf_group.v/mV, inf_group.m_inf, label=r'$m_\infty$')
    ax.plot(inf_group.v/mV, inf_group.n_inf, label=r'$n_\infty$')
    ax.plot(inf_group.v/mV, inf_group.h_inf, label=r'$h_\infty$')
    ax.set_xlabel('$v$ (mV)')
    ax.set_ylabel('steady-state activation')
    ax.yaxis.set_label_position("right")
    ax.yaxis.tick_right()
    ax.legend()


def plot_membrane_voltage(ax, statemon):
    """Plot simulation result: membrane potential.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with v recorded)
    """

    ax.plot(statemon.t/ms, statemon.v[0]/mV, label='membrane voltage')
    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel('$v$ (mV)')
    ax.axhline(0, linestyle='dashed')
    ax.legend()


def plot_gating_variable_activations(ax, statemon):
    """Plot simulation result: gating variables.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with m, n and h recorded)
    """

    ax.plot(statemon.t/ms, statemon.m[0], label='$m$')
    ax.plot(statemon.t/ms, statemon.n[0], label='$n$')
    ax.plot(statemon.t/ms, statemon.h[0], label='$h$')
    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel('activation')
    ax.legend()


def plot_conductances(ax, statemon):
    """Plot simulation result: conductances.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with g_K and g_Na recorded)
    """

    ax.plot(statemon.t/ms, statemon.g_K[0] / (mS/(cmeter**2)),
            label=r'$g_\mathregular{K}$')

    ax.plot(statemon.t/ms, statemon.g_Na[0] / (mS/(cmeter**2)),
            label=r'$g_\mathregular{Na}$')

    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel('$g$ (mS/cm$^2$)')
    ax.legend()


def plot_currents(ax, statemon):
    """Plot simulation result: currents.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with I_K, I_Na and I_L recorded)
    """

    ax.plot(statemon.t/ms,
            statemon.I_K[0] / (uA/(cmeter**2)),
            label=r'$I_\mathregular{K}$')

    ax.plot(statemon.t/ms, statemon.I_Na[0] / (uA/(cmeter**2)),
            label=r'$I_\mathregular{Na}$')

    ax.plot(statemon.t/ms, (statemon.I_Na[0] + statemon.I_K[0] +
                            statemon.I_L[0]) / (uA/(cmeter**2)),
            label=r'$I_\mathregular{Na} + I_\mathregular{K} + I_\mathregular{L}$')

    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel(r'I ($\mu$A/cm$^2$)')
    ax.legend()


def plot_current_stimulus(ax, statemon):
    """Plot simulation result: external current stimulus.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with I_ext recorded)
    """

    ax.plot(statemon.t/ms, statemon.I_ext[0] /
            (uA/(cmeter**2)), label=r'$I_\mathregular{ext}$')

    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel(r'I ($\mu$A/cm$^2$)')
    ax.legend()


def plot_gating_variable_time_constants(ax, statemon):
    """Plot simulation result: gating variable time constants.

    ax -- matplotlib axes to be plotted on
    statemon -- StateMonitor (with tau_m, tau_n and tau_h recorded)
    """

    ax.plot(statemon.t/ms, statemon.tau_m[0]/ms, label=r'$\tau_m$')
    ax.plot(statemon.t/ms, statemon.tau_n[0]/ms, label=r'$\tau_n$')
    ax.plot(statemon.t/ms, statemon.tau_h[0]/ms, label=r'$\tau_h$')

    ax.set_xlabel('$t$ (ms)')
    ax.set_ylabel(r'$\tau$ (ms)')
    ax.legend()


def run_simulation(parameters):
    """Run the simulation.

    parameters -- dictionary with parameters
    """

    equations = []
    for gating_variable in ["m", "n", "h"]:
        equations.append(
            construct_gating_variable_inf_equation(gating_variable))
        equations.append(
            construct_gating_variable_tau_equation(gating_variable))
        equations.append(construct_gating_variable_ode(gating_variable))
    equations += construct_neuron_ode()

    eqs_HH = reduce(operator.add, equations)
    group = NeuronGroup(1, eqs_HH, method='euler', namespace=parameters)

    group.v = parameters["v_initial"]

    group.m = parameters["m_initial"]
    group.n = parameters["n_initial"]
    group.h = parameters["h_initial"]

    statemon = StateMonitor(group, ['v',
                                    'I_ext',
                                    'm', 'n', 'h',
                                    'g_K', 'g_Na',
                                    'I_K', 'I_Na', 'I_L',
                                    'tau_m', 'tau_n', 'tau_h'],
                            record=True)

    defaultclock.dt = parameters["defaultclock_dt"]
    run(parameters["duration"])

    return statemon


def main(parameters):
    """Run simulation and return matplotlib figure.

    parameters -- dictionary with parameters
    """

    statemon = run_simulation(parameters)

    fig = plt.figure(figsize=(20, 15), constrained_layout=True)
    gs = fig.add_gridspec(6, 2)

    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[1, 0])
    ax2 = fig.add_subplot(gs[2, 0])
    ax3 = fig.add_subplot(gs[3, 0])
    ax4 = fig.add_subplot(gs[4, 0])
    ax5 = fig.add_subplot(gs[5, 0])
    ax6 = fig.add_subplot(gs[:3, 1])
    ax7 = fig.add_subplot(gs[3:, 1])

    plot_membrane_voltage(ax0, statemon)
    plot_gating_variable_activations(ax1, statemon)
    plot_conductances(ax2, statemon)
    plot_currents(ax3, statemon)
    plot_current_stimulus(ax4, statemon)
    plot_gating_variable_time_constants(ax5, statemon)

    plot_tau(ax6, parameters)
    plot_inf(ax7, parameters)

    return fig


parameters = {

    # Boltzmann function parameters
    'v_n_half': 12*mV,
    'v_m_half': 25*mV,
    'v_h_half': 3*mV,

    'k_n': 15*mV,
    'k_m': 9*mV,
    'k_h': -7*mV,

    # Gaussian function parameters
    'v_n_max': -14*mV,
    'v_m_max': 27*mV,
    'v_h_max': -2*mV,

    'sigma_n': 50*mV,
    'sigma_m': 30*mV,
    'sigma_h': 20*mV,

    'c_n_amp': 4.7*ms,
    'c_m_amp': 0.46*ms,
    'c_h_amp': 7.4*ms,

    'c_n_base': 1.1*ms,
    'c_m_base': 0.04*ms,
    'c_h_base': 1.2*ms,

    # conductances
    'g_K_bar': 36*mS / (cmeter**2),
    'g_Na_bar': 120*mS / (cmeter**2),
    'g_L': 0.3*mS / (cmeter**2),

    # reversal potentials
    'e_K': -12*mV,
    'e_Na': 120*mV,
    'e_L': 10.6*mV,

    # membrane capacitance
    'C_mem': 1*uF / cmeter**2,

    # initial membrane voltage
    'v_initial': 0*mV,

    # initial gating variable activations
    'm_initial': 0.05,
    'n_initial': 0.32,
    'h_initial': 0.60,

    # external stimulus at 2 ms with 4 uA/cm^2 and at 10 ms with 15 uA/cm^2
    # for 0.5 ms each
    'I_stim': TimedArray(values=([0]*4+[4]+[0]*15+[15]+[0])*uA/(cmeter**2),
                         dt=0.5*ms),

    # simulation time step
    'defaultclock_dt': 0.01*ms,

    # simulation duration
    'duration': 20*ms
}

linestyle_cycler = cycler('linestyle',['-','--',':','-.'])
plt.rc('axes', prop_cycle=linestyle_cycler)

fig = main(parameters)

plt.show()



"""
Utility functions for parsing expressions and statements.
"""

import re
from collections import Counter

import sympy
from sympy.printing.precedence import precedence
from sympy.printing.str import StrPrinter

from brian2.core.functions import DEFAULT_CONSTANTS, DEFAULT_FUNCTIONS, Function
from brian2.parsing.rendering import SympyNodeRenderer
from brian2.utils.caching import cached


def check_expression_for_multiple_stateful_functions(expr, variables):
    identifiers = re.findall(r"\w+", expr)
    # Don't bother counting if we don't have any duplicates in the first place
    if len(identifiers) == len(set(identifiers)):
        return
    identifier_count = Counter(identifiers)
    for identifier, count in identifier_count.items():
        var = variables.get(identifier, None)
        if count > 1 and isinstance(var, Function) and not var.stateless:
            raise NotImplementedError(
                f"The expression '{expr}' contains "
                f"more than one call of {identifier}. This "
                "is currently not supported since "
                f"{identifier} is a stateful function and "
                "its multiple calls might be "
                "treated incorrectly (e.g."
                "'rand() - rand()' could be "
                " simplified to "
                "'0.0')."
            )


def str_to_sympy(expr, variables=None):
    """
    Parses a string into a sympy expression. There are two reasons for not
    using `sympify` directly: 1) sympify does a ``from sympy import *``,
    adding all functions to its namespace. This leads to issues when trying to
    use sympy function names as variable names. For example, both ``beta`` and
    ``factor`` -- quite reasonable names for variables -- are sympy functions,
    using them as variables would lead to a parsing error. 2) We want to use
    a common syntax across expressions and statements, e.g. we want to allow
    to use `and` (instead of `&`) and function names like `ceil` (instead of
    `ceiling`).

    Parameters
    ----------
    expr : str
        The string expression to parse.
    variables : dict, optional
        Dictionary mapping variable/function names in the expr to their
        respective `Variable`/`Function` objects.

    Returns
    -------
    s_expr
        A sympy expression

    Raises
    ------
    SyntaxError
        In case of any problems during parsing.
    """
    if variables is None:
        variables = {}
    check_expression_for_multiple_stateful_functions(expr, variables)

    # We do the actual transformation in a separate function that is cached
    # If we cached `str_to_sympy` itself, it would also use the contents of the
    # variables dictionary as the cache key, while it is only used for the check
    # above and does not affect the translation to sympy
    return _str_to_sympy(expr)


@cached
def _str_to_sympy(expr):
    try:
        s_expr = SympyNodeRenderer().render_expr(expr)
    except (TypeError, ValueError, NameError) as ex:
        raise SyntaxError(f"Error during evaluation of sympy expression '{expr}': {ex}")

    return s_expr


class CustomSympyPrinter(StrPrinter):
    """
    Printer that overrides the printing of some basic sympy objects. E.g.
    print "a and b" instead of "And(a, b)".
    """

    def _print_And(self, expr):
        return " and ".join([f"({self.doprint(arg)})" for arg in expr.args])

    def _print_Or(self, expr):
        return " or ".join([f"({self.doprint(arg)})" for arg in expr.args])

    def _print_Not(self, expr):
        if len(expr.args) != 1:
            raise AssertionError(f'"Not" with {len(expr.args)} arguments?')
        return f"not ({self.doprint(expr.args[0])})"

    def _print_Relational(self, expr):
        return (
            f"{self.parenthesize(expr.lhs, precedence(expr))} "
            f"{self._relationals.get(expr.rel_op) or expr.rel_op} "
            f"{self.parenthesize(expr.rhs, precedence(expr))}"
        )

    def _print_Function(self, expr):
        # Special workaround for the int function
        if expr.func.__name__ == "int_":
            return f"int({self.stringify(expr.args, ', ')})"
        elif expr.func.__name__ == "Mod":
            return f"(({self.doprint(expr.args[0])})%({self.doprint(expr.args[1])}))"
        else:
            return f"{expr.func.__name__}({self.stringify(expr.args, ', ')})"


PRINTER = CustomSympyPrinter()


@cached
def sympy_to_str(sympy_expr):
    """
    sympy_to_str(sympy_expr)

    Converts a sympy expression into a string. This could be as easy as
    ``str(sympy_exp)`` but it is possible that the sympy expression contains
    functions like ``Abs`` (for example, if an expression such as
    ``sqrt(x**2)`` appeared somewhere). We do want to re-translate ``Abs`` into
    ``abs`` in this case.

    Parameters
    ----------
    sympy_expr : sympy.core.expr.Expr
        The expression that should be converted to a string.

    Returns
    str_expr : str
        A string representing the sympy expression.
    """
    # replace the standard functions by our names if necessary
    replacements = {
        f.sympy_func: sympy.Function(name)
        for name, f in DEFAULT_FUNCTIONS.items()
        if f.sympy_func is not None
        and isinstance(f.sympy_func, sympy.FunctionClass)
        and str(f.sympy_func) != name
    }
    # replace constants with our names as well
    replacements.update(
        {
            c.sympy_obj: sympy.Symbol(name)
            for name, c in DEFAULT_CONSTANTS.items()
            if str(c.sympy_obj) != name
        }
    )

    # Replace the placeholder argument by an empty symbol
    replacements[sympy.Symbol("_placeholder_arg")] = sympy.Symbol("")
    atoms = sympy_expr.atoms() | {f.func for f in sympy_expr.atoms(sympy.Function)}
    for old, new in replacements.items():
        if old in atoms:
            sympy_expr = sympy_expr.subs(old, new)
    expr = PRINTER.doprint(sympy_expr)

    return expr


def expression_complexity(expr, complexity=None):
    """
    Returns the complexity of an expression (either string or sympy)

    The complexity is defined as 1 for each arithmetic operation except divide which is 2,
    and all other operations are 20. This can be overridden using the complexity
    argument.

    Note: calling this on a statement rather than an expression is likely to lead to errors.

    Parameters
    ----------
    expr: `sympy.Expr` or str
        The expression.
    complexity: None or dict (optional)
        A dictionary mapping expression names to their complexity, to overwrite default behaviour.

    Returns
    -------
    complexity: int
        The complexity of the expression.
    """
    if isinstance(expr, str):
        # we do this because sympy.count_ops doesn't handle inequalities (TODO: handle sympy as well str)
        for op in ["<=", ">=", "==", "<", ">"]:
            expr = expr.replace(op, "+")
        # work around bug with rand() and randn() (TODO: improve this)
        expr = expr.replace("rand()", "rand(0)")
        expr = expr.replace("randn()", "randn(0)")
    subs = {"ADD": 1, "DIV": 2, "MUL": 1, "SUB": 1}
    if complexity is not None:
        subs.update(complexity)
    ops = sympy.count_ops(expr, visual=True)
    for atom in ops.atoms():
        if hasattr(atom, "name"):
            subs[atom.name] = 20  # unknown operations assumed to have a large cost
    return ops.evalf(subs=subs)



__all__ = [
    "FeatureTest",
    "SpeedTest",
    "InaccuracyError",
    "Configuration",
    "run_feature_tests",
]
# isort: skip_file # We need to do the base import first to prevent a circular import later
from .base import *
from . import input, monitors, neurongroup, speed, synapses



#!/usr/bin/env python3
"""
Reproduces Figure 2F of

The Use of Hebbian Cell Assemblies for Nonlinear Computation
by Tetzlaff C., Dasgupta S., Kulvicius T. and Wörgötter F.

Sci Rep 5, 12866 (2015).
https://doi.org/10.1038/srep12866

Sebastian Schmitt, 2022
"""

import numpy as np
import matplotlib.pyplot as plt

from brian2 import NeuronGroup, Synapses, StateMonitor, run, defaultclock, ms, second, TimedArray, seed

# random seed that gives curves similar to the ones in the publication
seed(9873487)

# neuron parameters (sigmoidal activation)
beta = 0.03
epsilon = 120
F_max = 100
F_T = 1
tau_u = 1*ms
R = 0.012

# plasticity timescales
tau_ratio = 60
# hebbian
tau_H = 3e4*ms
# synaptic scaling
tau_SS = tau_ratio * tau_H

# synaptic weights
W_max = np.sqrt(tau_ratio*(F_max**2/(F_max - F_T)))
W_ext = W_max
W_input = W_max
W_I = 0.3*W_max

# stimulus
N_units = 100
N_stim_units = 20
stim_A_units_until = N_stim_units
stim_B_units_from = N_units-N_stim_units

# connection probabilities
p_E = 0.1
p_I = 0.2

# paper uses 0.3*ms
DT = 0.5*ms
defaultclock.dt = DT

# duration of a learning trial
lt = 5000*DT

duration = 100*lt
no_input_until = 5*lt
balanced_until = duration/2

# gate balanced presentation of stimulus 1 and 2
balanced = TimedArray([lt_counter*lt < balanced_until for lt_counter in range(int(duration/lt))], dt=lt)

# function used for stimulus (typo in paper, +1 is not part of the argument of sin)
stim_func = TimedArray([100*(np.sin(0.1*(i+1))+1) for i in range(int(duration/DT))], dt=DT)

# gate learning phase of either stimulus 1 or 2
learning_phase = TimedArray([i%10 > 3 for i in range(int(duration/(0.1*lt)))], dt=0.1*lt)

# if not balanced present stimulus A three times more often than stimulus B
stim_A_gate = TimedArray([lt_counter % 2 == 0 if balanced(lt_counter*lt) else lt_counter % 4 in [0,1,2]
                          for lt_counter in range(int(duration/lt))], dt=lt)

stim_B_gate = TimedArray([lt_counter % 2 == 1 if balanced(lt_counter*lt) else lt_counter % 4 == 3
                          for lt_counter in range(int(duration/lt))], dt=lt)

# noise is applied also during stimulation
neurons = NeuronGroup(N_units,
                      """
                      F = F_max/(1+exp(beta*(epsilon-u))) : 1
                      du/dt = (-u + R*(I_E - I_I + W_input*(I_stim_A + I_stim_B)))/tau_u + R*W_ext*20*sqrt((DT/ms)/ms)*xi: 1
                      I_E : 1
                      I_I : 1
                      index : 1 (constant)
                      stim_units_A = index < stim_A_units_until :  boolean
                      stim_units_B = index >= (stim_B_units_from) : boolean
                      I_stim_A = learning_phase(t)*int(stim_units_A)*stim_A_gate(t)*stim_func(t) : 1
                      I_stim_B = learning_phase(t)*int(stim_units_B)*stim_B_gate(t)*stim_func(t) : 1
                      """,
                      method = "euler")
neurons.index = range(len(neurons))

# excitatory connections with Hebbian plasticity and synaptic scaling
synapses_E = Synapses(neurons, neurons,
                      """
                      dw/dt = 1/tau_H*F_pre*F_post + 1/tau_SS*(F_T - F_post)*w**2 : 1 (clock-driven)
                      I_E_post = w*F_pre : 1 (summed)
                      """,
                      method="euler"
                      )
# do not connect between the two populations of stimulated neurons
synapses_E.connect(p=p_E, condition="((j > stim_A_units_until and i >= stim_B_units_from) or (j < stim_B_units_from and i < stim_A_units_until))"
                                    "or ((i > stim_A_units_until and i < stim_B_units_from) and (j > stim_A_units_until and j < stim_B_units_from))")

# fixed weight inhibitory connections
synapses_I = Synapses(neurons, neurons,
                      """
                      w : 1
                      I_I_post = w*F_pre : 1 (summed)
                      """
                      )
synapses_I.connect(p=p_I)
synapses_I.w = W_I

statemon_neurons = StateMonitor(neurons, ["F", "I_stim_A", "I_stim_B"], record=True, dt=100*defaultclock.dt)
statemon_synapses_E = StateMonitor(synapses_E, "w", record=True, dt=100*defaultclock.dt)
statemon_synapses_for_assembly_analysis = StateMonitor(synapses_E, "w", record=True, dt=lt)

run(duration, report="text")

# threshold saying that synaptic efficacies larger than theta are
# 'strong' and others are 'weak'
theta = 0.5*W_max

in_assembly_A = []
in_assembly_B = []

# traverse through the graph following 'strong' synapses
def go(W, source, units_in_assembly):
    units_in_assembly.add(source)
    # check all possible targets
    for target in range(N_units):
        w = W[source][target]
        if w > theta:
            W[source][target] = 0
            go(W, target, units_in_assembly)

# for each learning trial
for ws in statemon_synapses_for_assembly_analysis.w.T:

    # construct a full weight matrix
    W = np.full((N_units, N_units), np.nan)
    W[synapses_E.i[:], synapses_E.j[:]] = ws

    for in_assembly, stim_units in zip([in_assembly_A, in_assembly_B],
                                       [range(stim_A_units_until),
                                        range(stim_B_units_from, N_units)]):

        units_in_assembly = set()

        # start with units that are stimulated
        for stim_unit in stim_units:
            go(W, stim_unit, units_in_assembly)

        in_assembly.append(len(units_in_assembly))

# competitive development of the two competing cell assemblies A and B as a function of the input protocol
fig, ax = plt.subplots()

ax.plot(in_assembly_A, linestyle="None", marker='o', color='orange', label="A")
ax.plot(in_assembly_B, linestyle="None", marker='o', color='olivedrab', label="B")
ax.set_ylim(19, 51)
ax.set_xlim(0, 100)

ax.set_ylabel("Neurons in Cell Assembly [%]")
ax.set_xlabel("Learning Trial")

ax.axvline(balanced_until/lt, linestyle='dashed', color='k')

ax.text(15, 52, " A A", color='orange', fontfamily="monospace", fontsize="xx-large")
ax.text(15, 52, "  B B", color='olivedrab', fontfamily="monospace", fontsize="xx-large")

ax.text(65, 52, " 3A 3A", color='orange', fontfamily="monospace", fontsize="xx-large")
ax.text(65, 52, "   B  B", color='olivedrab', fontfamily="monospace", fontsize="xx-large")

plt.show()

# stimulus, neuronal activity and excitatory weights as function of time
fig, axes = plt.subplots(3, sharex=True)

axes[0].plot(statemon_neurons.I_stim_A[0], label="A", color='orange')
axes[0].plot(statemon_neurons.I_stim_B[-1], label="B", color='olivedrab')
axes[0].legend(loc="upper right")
axes[0].set_title("Stimulus")

axes[1].imshow(statemon_neurons.F, aspect='auto')
axes[1].set_title("Neuron Activity")
axes[1].axhline(stim_A_units_until, linestyle='dashed', color='white')
axes[1].axhline(stim_B_units_from, linestyle='dashed', color='white')

axes[2].imshow(statemon_synapses_E.w, aspect='auto')
axes[2].set_title("Excitatory Weights")

axes[2].set_xticks(range(0, 5000, 250))
axes[2].set_xticklabels(f"{i}" for i in range(0, 100, 5))

axes[2].set_xlabel("Learning Trial")
axes[2].set_xlim(0, 5000)

fig.tight_layout()

plt.show()



"""
Package providing import/export support.
"""

from .dictlike import *
from .importexport import *

__all__ = ["ImportExport"]

# Register the standard ImportExport methods
for obj in [DictImportExport(), PandasImportExport()]:
    ImportExport.register(obj)



"""
Module containing the Cython CodeObject for code generation for integration using the ODE solver provided in the
GNU Scientific Library (GSL)
"""

import os
import sys
from distutils.errors import CompileError

from brian2.core.preferences import prefs

from ...generators.cython_generator import CythonCodeGenerator
from ...generators.GSL_generator import GSLCythonCodeGenerator
from ..cython_rt import CythonCodeObject

__all__ = ["GSLCythonCodeObject", "IntegrationError"]


class GSLCompileError(Exception):
    pass


class IntegrationError(Exception):
    """
    Error used to signify that GSL was unable to complete integration (only works for cython)
    """

    pass


class GSLCythonCodeObject(CythonCodeObject):
    templater = CythonCodeObject.templater.derive("brian2.codegen.runtime.GSLcython_rt")

    # CodeGenerator that is used to do bulk of abstract_code --> language specific translation
    original_generator_class = CythonCodeGenerator
    generator_class = GSLCythonCodeGenerator

    def compile(self):
        self.libraries += ["gsl", "gslcblas"]
        self.headers += [
            "<stdio.h>",
            "<stdlib.h>",
            "<gsl/gsl_odeiv2.h>",
            "<gsl/gsl_errno.h>",
            "<gsl/gsl_matrix.h>",
        ]
        if sys.platform == "win32":
            self.define_macros += [("WIN32", "1"), ("GSL_DLL", "1")]
        if prefs.GSL.directory is not None:
            self.include_dirs += [prefs.GSL.directory]
            self.library_dirs += [
                os.path.abspath(os.path.join(prefs.GSL.directory, "..", "lib"))
            ]
            if sys.platform == "win32":
                GSL_bin = os.path.abspath(
                    os.path.join(os.path.join(prefs.GSL.directory, "..", "bin"))
                )
                os.add_dll_directory(GSL_bin)
        try:
            super().compile()
        except CompileError as err:
            raise GSLCompileError(
                "\nCompilation of files generated for integration with GSL has failed."
                "\nOne cause for this could be incorrect installation of GSL itself."
                "\nIf GSL is installed but Python cannot find the correct files, it is "
                "also possible to give the gsl directory manually by specifying "
                "prefs.GSL.directory = ..."
            ) from err



"""
File system tools
"""

import os

__all__ = [
    "ensure_directory",
    "ensure_directory_of_file",
    "in_directory",
    "copy_directory",
]


def ensure_directory_of_file(f):
    """
    Ensures that a directory exists for filename to go in (creates if
    necessary), and returns the directory path.
    """
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)
    return d


def ensure_directory(d):
    """
    Ensures that a given directory exists (creates it if necessary)
    """
    if not os.path.exists(d):
        os.makedirs(d)
    return d


class in_directory:
    """
    Safely temporarily work in a subdirectory

    Usage::

        with in_directory(directory):
            ... do stuff here

    Guarantees that the code in the with block will be executed in directory,
    and that after the block is completed we return to the original directory.
    """

    def __init__(self, new_dir):
        self.orig_dir = os.getcwd()
        self.new_dir = new_dir

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *exc_info):
        os.chdir(self.orig_dir)


def copy_directory(source, target):
    """
    Copies directory source to target.
    """
    relnames = []
    sourcebase = os.path.normpath(source) + os.path.sep
    for root, _, filenames in os.walk(source):
        for filename in filenames:
            fullname = os.path.normpath(os.path.join(root, filename))
            relname = fullname.replace(sourcebase, "")
            relnames.append(relname)
            tgtname = os.path.join(target, relname)
            ensure_directory_of_file(tgtname)
            with open(fullname) as f:
                contents = f.read()
            if os.path.exists(tgtname):
                with open(tgtname) as f:
                    if f.read() == contents:
                        continue
            with open(tgtname, "w") as f:
                f.write(contents)
    return relnames



# from https://gist.github.com/schlamar/2311116
import os
import sys
import traceback
from functools import wraps
from multiprocessing import Process, Queue


def processify(func):
    '''Decorator to run a function as a process.
    Be sure that every argument and the return value
    is *pickable*.
    The created process is joined, so the code does not
    run in parallel.
    '''

    def process_func(q, *args, **kwargs):
        try:
            ret = func(*args, **kwargs)
        except Exception:
            ex_type, ex_value, tb = sys.exc_info()
            error = ex_type, ex_value, ''.join(traceback.format_tb(tb))
            ret = None
        else:
            error = None

        q.put((ret, error))

    # register original function with different name
    # in sys.modules so it is pickable
    process_func.__name__ = func.__name__ + 'processify_func'
    setattr(sys.modules[__name__], process_func.__name__, process_func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        q = Queue()
        p = Process(target=process_func, args=[q] + list(args), kwargs=kwargs)
        p.start()
        p.join()
        ret, error = q.get()

        if error:
            ex_type, ex_value, tb_str = error
            message = '%s (in subprocess)\n%s' % (ex_value.message, tb_str)
            raise ex_type(message)

        return ret
    return wrapper


@processify
def test_function():
    return os.getpid()


@processify
def test_exception():
    raise RuntimeError('xyz')


def test():
    print(os.getpid())
    print(test_function())
    test_exception()

if __name__ == '__main__':
    test()




import os
import shutil
import glob
import codecs

from nbformat import NotebookNode
from nbformat.v4 import reads
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.exporters.notebook import NotebookExporter
from nbconvert.exporters.rst import RSTExporter

from brian2.utils.stringtools import deindent, indent


src_dir = os.path.abspath('../../../tutorials')
target_dir = os.path.abspath('../../../docs_sphinx/resources/tutorials')

# Start from scratch to avoid left-over files due to renamed tutorials, changed
# cell numbers, etc.
if os.path.exists(target_dir):
    shutil.rmtree(target_dir)
os.mkdir(target_dir)

tutorials = []
for fname in sorted(glob.glob1(src_dir, '*.ipynb')):
    basename = fname[:-6]
    output_ipynb_fname = os.path.join(target_dir, fname)
    output_rst_fname = os.path.join(target_dir, basename + '.rst')

    print('Running', fname)
    with open(os.path.join(src_dir, fname), 'r') as f:
        notebook = reads(f.read())

    # The first line of the tutorial file should give the title
    title = notebook.cells[0]['source'].split('\n')[0].strip('# ')
    tutorials.append((basename, title))

    # Execute the notebook
    preprocessor = ExecutePreprocessor(timeout=None)
    preprocessor.allow_errors = True
    notebook, _ = preprocessor.preprocess(notebook,
                                          {'metadata': {'path': src_dir}})

    print('Saving notebook and converting to RST')
    exporter = NotebookExporter()
    output, _ = exporter.from_notebook_node(notebook)
    with codecs.open(output_ipynb_fname, 'w', encoding='utf-8') as f:
        f.write(output)

    # Insert a note about ipython notebooks with a download link
    note = deindent('''
    .. only:: html

        .. |launchbinder| image:: http://mybinder.org/badge.svg
        .. _launchbinder: https://mybinder.org/v2/gh/brian-team/brian2-binder/master?filepath=tutorials/{tutorial}.ipynb
    
        .. note::
           This tutorial is a static non-editable version. You can launch an
           interactive, editable version without installing any local files
           using the Binder service (although note that at some times this
           may be slow or fail to open): |launchbinder|_
    
           Alternatively, you can download a copy of the notebook file
           to use locally: :download:`{tutorial}.ipynb`
    
           See the :doc:`tutorial overview page <index>` for more details.

    '''.format(tutorial=basename))
    notebook.cells.insert(1, NotebookNode(cell_type='raw', metadata={},
                                          source=note))
    exporter = RSTExporter()
    output, resources = exporter.from_notebook_node(notebook,
                                                    resources={'unique_key': basename+'_image'})
    with codecs.open(output_rst_fname, 'w', encoding='utf-8') as f:
        f.write(output)

    for image_name, image_data in resources['outputs'].items():
        with open(os.path.join(target_dir, image_name), 'wb') as f:
            f.write(image_data)

print('Generating index.rst')

text = '''
..
    This is a generated file, do not edit directly.
    (See dev/tools/docs/build_tutorials.py)

Tutorials
=========

The tutorial consists of a series of `Jupyter Notebooks`_ [#]_.

.. only:: html

    You can quickly view these using the first links below. To use them interactively - allowing you
    to edit and run the code - there are two options. The easiest option is to click
    on the "Launch Binder" link, which will open up an interactive version in the
    browser without having to install Brian locally. This uses the
    `mybinder.org <http://mybinder.org>`_ service. Occasionally, this
    service will be down or running slowly. The other option is to download the
    notebook file and run it locally, which requires you to have Brian installed.

For more information about how to use Jupyter Notebooks, see the
`Jupyter Notebook documentation`_.

.. toctree::
   :maxdepth: 1
   :titlesonly:

'''
for tutorial, _ in tutorials:
    text += '   ' + tutorial + '\n'
text += '''
.. only:: html

    Interactive notebooks and files
    -------------------------------
'''
for tutorial, _ in tutorials:
    text += indent(deindent('''
    .. |launchbinder{tutid}| image:: http://mybinder.org/badge.svg
    .. _launchbinder{tutid}: https://mybinder.org/v2/gh/brian-team/brian2-binder/master?filepath=tutorials/{tutorial}.ipynb
    '''.format(tutorial=tutorial, tutid=tutorial.replace('-', ''))))

text += '\n'
for tutorial, title in tutorials:
    text += '    * |launchbinder{tutid}|_ :download:`{title} <{tutorial}.ipynb>`\n'.format(title=title,
                                                tutorial=tutorial, tutid=tutorial.replace('-', ''))
text += '''

.. _`Jupyter Notebooks`: http://jupyter-notebook-beginner-guide.readthedocs.org/en/latest/what_is_jupyter.html
.. _`Jupyter`: http://jupyter.org/
.. _`Jupyter Notebook documentation`: http://jupyter.readthedocs.org/

.. [#] Formerly known as "IPython Notebooks".
'''
with open(os.path.join(target_dir, 'index.rst'), 'w') as f:
    f.write(text)



#!/usr/bin/env python
"""
Neurons with gap junctions.
"""
from brian2 import *

n = 10
v0 = 1.05
tau = 10*ms

eqs = '''
dv/dt = (v0 - v + Igap) / tau : 1
Igap : 1 # gap junction current
'''

neurons = NeuronGroup(n, eqs, threshold='v > 1', reset='v = 0',
                      method='exact')
neurons.v = 'i * 1.0 / (n-1)'
trace = StateMonitor(neurons, 'v', record=[0, 5])

S = Synapses(neurons, neurons, '''
             w : 1 # gap junction conductance
             Igap_post = w * (v_pre - v_post) : 1 (summed)
             ''')
S.connect()
S.w = .02

run(500*ms)

plot(trace.t/ms, trace[0].v)
plot(trace.t/ms, trace[5].v)
xlabel('Time (ms)')
ylabel('v')
show()



{# USES_VARIABLES { _synaptic_pre, _synaptic_post, sources, targets, N,
                    N_pre, N_post, _source_offset, _target_offset }
#}
{# WRITES_TO_READ_ONLY_VARIABLES { _synaptic_pre, _synaptic_post, N}
#}
{% extends 'common_group.pyx' %}

{% block maincode %}
    cdef size_t _old_num_synapses = {{N}}
    cdef size_t _new_num_synapses = _old_num_synapses + _num{{sources}}

    {{_dynamic__synaptic_pre}}.resize(_new_num_synapses)
    {{_dynamic__synaptic_post}}.resize(_new_num_synapses)
    # Get the potentially newly created underlying data arrays
    cdef int32_t[:] _synaptic_pre_data = {{_dynamic__synaptic_pre}}.data
    cdef int32_t[:] _synaptic_post_data = {{_dynamic__synaptic_post}}.data

    for _idx in range(_num{{sources}}):
        # After this code has been executed, the arrays _real_sources and
        # _real_variables contain the final indices. Having any code here it all is
        # only necessary for supporting subgroups
        {{ vector_code | autoindent }}
        _synaptic_pre_data[_idx + _old_num_synapses] = _real_sources
        _synaptic_post_data[_idx + _old_num_synapses] = _real_targets
    
    # now we need to resize all registered variables and set the total number
    # of synapses (via Python)
    _owner._resize(_new_num_synapses)

    # And update N_incoming, N_outgoing and synapse_number
    _owner._update_synapse_numbers(_old_num_synapses)
{% endblock %}



"""
Unsupervised learning using STDP
--------------------------------
Diehl, P. U., & Cook, M. (2015). Unsupervised learning of digit
recognition using spike-timing-dependent plasticity. Frontiers in
computational neuroscience, 9, 99.

This script replicates the small 2x400-model. It has no command line
parameters. Instead, you control it by changing the constants below
the imports. Run the script with MODE set to "train" which
(eventually) creates the files theta.npy and weights.npy in the
DATA_PATH directory. Rerun it with MODE set to "observe" to create the
assign.npy file in the same directory. Then, run "test" to create a
confusion matrix in confusion.npy. Finally, you can use "plot" to
plot the confusion matrix. The script also creates a few auxilliary
.npy files useful for analysis. The script requires the progressbar2
library.

MNIST_PATH should point to the directory storing the unzipped *-byte
MNIST files (e.g. from https://github.com/cvdfoundation/mnist).
For reasonable accuracy, N_TRAIN should be 50,000+ and N_OBSERVE 1,000+.

Written in 2024 by Björn A. Lindqvist <bjourne@gmail.com>
"""
from brian2 import *
from collections import defaultdict
from pathlib import Path
from progressbar import progressbar
from random import randrange, seed as rseed
from struct import unpack
import numpy as np

# Switch between "train", "observe", and "test" to tune parameters,
# observe excitatory spiking, and test accuracy, respectively.
# Use "plot" to plot the confusion matrix.
MODE = 'test'

# Number of training, observation, and testing samples
N_TRAIN = 25_000
N_OBSERVE = 2_000
N_TEST = 1_000

# Random seed value
SEED = 42

# Storage paths
MNIST_PATH = Path('../mnist')
DATA_PATH = Path('data')

# Number of weight save points
N_SAVE_POINTS = 100

# Don't change these values unless you know what you're doing.
N_INP = 784
N_NEURONS = 400
V_EXC_REST = -65 * mV
V_INH_REST = -60 * mV
INTENSITY = 2

# Weights of exc->inh and inh->exc synapses
W_EXC_INH = 10.4
W_INH_EXC = 17.0

def save_npy(arr, path):
    arr = np.array(arr)
    print('%-9s %-15s => %-30s' % ('Saving', arr.shape, path))
    np.save(path, arr)

def load_npy(path):
    arr = np.load(path)
    print('%-9s %-30s => %-15s' % ('Loading', path, arr.shape))
    return arr

def read_mnist(training):
    tag = 'train' if training else 't10k'
    images = open(MNIST_PATH / ('%s-images-idx3-ubyte' % tag), 'rb')
    images.read(4)
    n_images = unpack('>I', images.read(4))[0]
    n_rows = unpack(">I", images.read(4))[0]
    n_cols = unpack(">I", images.read(4))[0]
    
    labels = open(MNIST_PATH / ('%s-labels-idx1-ubyte' % tag), 'rb')
    labels.read(4)
    x = np.frombuffer(images.read(), dtype = np.uint8)
    x = x.reshape(n_images, -1) / 8.0
    y = np.frombuffer(labels.read(), dtype = np.uint8)
    return x, y

def build_network(training):
    eqs = '''
    dv/dt = (v_rest - v + i_exc + i_inh) / tau_mem  : volt (unless refractory)
    i_exc = ge * -v                         : volt
    i_inh = gi * (v_inh_base - v)           : volt
    dge/dt = -ge/(1 * ms)                   : 1
    dgi/dt = -gi/(2 * ms)                   : 1
    dtimer/dt = 1                           : second
    '''
    reset = 'v = %r; timer = 0 * ms' % V_EXC_REST
    if training:
        exc_eqs = eqs + '''
        dtheta/dt = -theta / (1e7 * ms)         : volt
        '''
        arr_theta = np.ones(N_NEURONS) * 20 * mV
        reset += '; theta += 0.05 * mV'
    else:
        exc_eqs = eqs + '''
        theta                                   : volt
        '''
        arr_theta = load_npy(DATA_PATH / 'theta.npy') * volt
    exc_eqs = Equations(exc_eqs,
                        tau_mem = 100 * ms,
                        v_rest = V_EXC_REST,
                        v_inh_base = -100 * mV)
    # Note that this neuron has a bit of un unusual refractoriness mechanism:
    # The membrane potential is clamped for 5ms, but spikes are prevented for 50ms
    # This has been taken from the original code.
    ng_exc = NeuronGroup(
        N_NEURONS, exc_eqs,
        threshold = 'v > (theta - 72 * mV) and (timer > 50 * ms)',
        refractory = 5 * ms,
        reset = reset,
        method = 'euler',
        name = 'exc')
    ng_exc.v = V_EXC_REST
    ng_exc.theta = arr_theta

    inh_eqs = Equations(eqs,
                        tau_mem = 10 * ms,
                        v_rest = V_INH_REST,
                        v_inh_base = -85 * mV)
    ng_inh = NeuronGroup(N_NEURONS, inh_eqs,
                         threshold = 'v > -40 * mV',
                         refractory = 2 * ms,
                         reset = 'v = -45 * mV',
                         method = 'euler',
                         name = 'inh')
    ng_inh.v = V_INH_REST

    syns_exc_inh = Synapses(ng_exc, ng_inh,
                            on_pre = 'ge_post += %f' % W_EXC_INH)
    syns_exc_inh.connect(j = 'i')

    syns_inh_exc = Synapses(ng_inh, ng_exc,
                            on_pre = 'gi_post += %f' % W_INH_EXC)
    syns_inh_exc.connect("i != j")

    pg_inp = PoissonGroup(N_INP, 0 * Hz, name = 'inp')

    # During training, inp->exc synapse weights are plastic.
    model = 'w : 1'
    on_post = ''
    on_pre = 'ge_post += w'
    if training:
        on_pre += '; pre = 1.; w = clip(w - 0.0001 * post1, 0, 1.0)'
        on_post += 'post2bef = post2; w = clip(w + 0.01 * pre * post2bef, 0, 1.0); post1 = 1.; post2 = 1.'
        model += '''
        post2bef                        : 1
        dpre/dt   = -pre/(20 * ms)      : 1 (event-driven)
        dpost1/dt = -post1/(20 * ms)    : 1 (event-driven)
        dpost2/dt = -post2/(40 * ms)    : 1 (event-driven)
        '''
        weights = (np.random.random(N_INP * N_NEURONS) + 0.01) * 0.3
    else:
        weights = load_npy(DATA_PATH / 'weights.npy')

    syns_inp_exc = Synapses(
        pg_inp, ng_exc,
        model = model,
        on_pre = on_pre,
        on_post = on_post,
        name = 'inp_exc'
    )
    syns_inp_exc.connect(True)
    syns_inp_exc.delay = 'rand() * 10 * ms'
    syns_inp_exc.w = weights

    exc_mon = SpikeMonitor(ng_exc, name = 'sp_exc')
    net = Network([pg_inp, ng_exc, ng_inh,
                   syns_inp_exc, syns_exc_inh, syns_inh_exc,
                   exc_mon])
    # Initialize
    net.run(0 * ms)
    return net

def show_sample(net, sample, intensity):
    exc_mon = net['sp_exc']
    prev = exc_mon.count[:]
    net['inp'].rates = sample * intensity * Hz
    net.run(350 * ms)
    # Don't count spikes occuring during the 150 ms rest.
    next = exc_mon.count[:]
    net['inp'].rates = 0 * Hz
    net.run(150 * ms)
    pat = next - prev
    cnt = np.sum(pat)
    if cnt < 5:
        return show_sample(net, sample, intensity + 1)
    return pat

def predict(groups, rates):
    return np.argmax([rates[grp].mean() for grp in groups])

def test():
    conf = np.zeros((10, 10))
    assign = np.load(DATA_PATH / 'assign.npy')
    groups = [np.where(assign == i)[0] for i in range(10)]

    X, Y = read_mnist(False)
    net = build_network(False)
    for i in progressbar(range(N_TEST)):
        ix = randrange(len(X))
        exc = show_sample(net, X[ix], INTENSITY)
        guess = predict(groups, exc)
        real = Y[ix]
        conf[real, guess] += 1

    print('Accuracy: %6.3f' % (np.trace(conf) / np.sum(conf)))
    conf = conf/conf.sum(axis=1)[:,None]
    print(np.around(conf, 2))
    save_npy(conf, DATA_PATH / 'confusion.npy')

def normalize_plastic_weights(syns):
    conns = np.reshape(syns.w, (N_INP, N_NEURONS))
    col_sums = np.sum(conns, axis = 0)
    factors = 78./ col_sums
    conns *= factors
    syns.w = conns.reshape(-1)

def stats(net):
    tick = defaultclock.timestep[:]
    cnt = np.sum(net['sp_exc'].count[:])

    inp_exc = net['inp_exc']
    w_mu = np.mean(inp_exc.w)
    w_std = np.std(inp_exc.w)

    exc = net['exc']
    theta = exc.theta / mV
    theta_mu = np.mean(theta)
    theta_sig = np.std(theta)
    return [tick, cnt, w_mu, w_std, theta_mu, theta_sig]

def train():
    X, Y = read_mnist(True)
    n_samples = X.shape[0]
    net = build_network(True)
    rows = [stats(net) + [-1]]
    w_hist = [np.array(net['inp_exc'].w)]

    ratio = max(N_TRAIN // N_SAVE_POINTS, 1)
    for i in progressbar(range(N_TRAIN)):
        ix = i % n_samples
        normalize_plastic_weights(net['inp_exc'])
        show_sample(net, X[ix], INTENSITY)
        rows.append(stats(net) + [Y[ix]])
        if i % ratio == 0:
            w_hist.append(np.array(net['inp_exc'].w))

    save_npy(rows, DATA_PATH / 'train_stats.npy')
    save_npy(w_hist, DATA_PATH / 'train_w_hist.npy')
    save_npy(net['inp_exc'].w, DATA_PATH / 'weights.npy')
    save_npy(net['exc'].theta, DATA_PATH / 'theta.npy')

def observe():
    X, Y = read_mnist(True)
    n_samples = X.shape[0]
    net = build_network(False)
    rows = [stats(net) + [-1]]
    responses = defaultdict(list)

    for i in progressbar(range(N_OBSERVE)):
        ix = i % n_samples
        sample = X[ix]
        cls = Y[ix]
        exc = show_sample(net, sample, INTENSITY)
        rows.append(stats(net) + [Y[ix]])
        responses[cls].append(exc)

    res = np.zeros((10, N_NEURONS))
    for cls, vals in responses.items():
        res[cls] = np.array(vals).mean(axis = 0)

    assign = np.argmax(res, axis = 0)
    save_npy(assign, DATA_PATH / 'assign.npy')
    save_npy(rows, DATA_PATH / 'observe_stats.npy')

def plot():
    conf = np.load(DATA_PATH / "confusion.npy")

    import matplotlib.pyplot as plt

    plt.imshow(100*conf, interpolation="nearest", cmap=plt.cm.Blues)
    for i, j in itertools.product(range(conf.shape[0]), range(conf.shape[1])):
        if conf[i, j] == 0:
            continue
        plt.text(
            j,
            i,
            f"{round(100*conf[i, j])}%",
            horizontalalignment="center",
            verticalalignment="center",
            color="white" if conf[i, j] > 0.5 else "black",
        )
    plt.colorbar()
    plt.xticks(range(10))
    plt.yticks(range(10))
    plt.xlabel("Predicted label")
    plt.ylabel("True label")
    plt.show()


if __name__ == '__main__':
    seed(SEED)
    rseed(SEED)
    DATA_PATH.mkdir(parents = True, exist_ok = True)
    cmds = dict(train=train, observe=observe, test=test, plot=plot)
    cmds[MODE]()



"""
Module containing fixtures and hooks used by the pytest test suite.
"""

import re

import numpy as np
import pytest

from brian2.core.clocks import defaultclock
from brian2.core.functions import DEFAULT_FUNCTIONS, Function
from brian2.devices import get_device, reinit_devices
from brian2.devices.device import reinit_and_delete, set_device
from brian2.units import ms


def pytest_ignore_collect(collection_path, config):
    if config.option.doctestmodules:
        if "tests" in collection_path.parts:
            return True  # Ignore tests package for doctests
    # Do not test brian2.hears bridge (needs Brian1)
    if collection_path.name == "hears.py":
        return True


# The "random" values are always 0.5
def fake_randn(vectorisation_idx):
    return 0.5 * np.ones_like(vectorisation_idx)


fake_randn = Function(
    fake_randn, arg_units=[], return_unit=1, auto_vectorise=True, stateless=False
)
fake_randn.implementations.add_implementation(
    "cpp",
    """
                                              double randn(int vectorisation_idx)
                                              {
                                                  return 0.5;
                                              }
                                              """,
)
fake_randn.implementations.add_implementation(
    "cython",
    """
                                    cdef double randn(int vectorisation_idx):
                                        return 0.5
                                    """,
)


@pytest.fixture
def fake_randn_randn_fixture():
    orig_randn = DEFAULT_FUNCTIONS["randn"]
    DEFAULT_FUNCTIONS["randn"] = fake_randn
    yield None
    DEFAULT_FUNCTIONS["randn"] = orig_randn


# Fixture that is used for all tests
@pytest.fixture(autouse=True)
def setup_and_teardown(request):
    # Set preferences before each test
    import brian2

    if hasattr(request.config, "workerinput"):
        config = request.config.workerinput
        for key, value in config["brian_prefs"].items():
            if isinstance(value, tuple) and value[0] == "TYPE":
                matches = re.match(r"<(type|class) 'numpy\.(.+)'>", value[1])
                if matches is None or len(matches.groups()) != 2:
                    raise TypeError(
                        f"Do not know how to handle {value[1]} in preferences"
                    )
                t = matches.groups()[1]
                if t == "float64":
                    value = np.float64
                elif t == "float32":
                    value = np.float32
                elif t == "int64":
                    value = np.int64
                elif t == "int32":
                    value = np.int32

            brian2.prefs[key] = value
        set_device(config["device"], **config["device_options"])
    else:
        for k, v in request.config.brian_prefs.items():
            brian2.prefs[k] = v
        set_device(request.config.device, **request.config.device_options)
    brian2.prefs._backup()
    # Print output changed in numpy 1.14, stick with the old format to
    # avoid doctest failures
    try:
        np.set_printoptions(legacy="1.13")
    except TypeError:
        pass  # using a numpy version < 1.14

    yield  # run test

    # Reset defaultclock.dt to be sure
    defaultclock.dt = 0.1 * ms


# (Optionally) mark tests raising NotImplementedError as skipped (mostly used
# for testing Brian2GeNN)
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    if hasattr(item.config, "workerinput"):
        fail_for_not_implemented = item.config.workerinput["fail_for_not_implemented"]
    else:
        fail_for_not_implemented = item.config.fail_for_not_implemented
    outcome = yield
    rep = outcome.get_result()
    if rep.outcome == "failed":
        project_dir = getattr(get_device(), "project_dir", None)
        if project_dir is not None:
            rep.sections.append(("Standalone project directory", f"{project_dir}"))
        reinit_devices()
        if not fail_for_not_implemented:
            exc_cause = getattr(call.excinfo.value, "__cause__", None)
            if call.excinfo.errisinstance(NotImplementedError) or isinstance(
                exc_cause, NotImplementedError
            ):
                rep.outcome = "skipped"
                r = call.excinfo._getreprcrash()
                rep.longrepr = (str(r.path), r.lineno, r.message)
    else:
        # clean up after the test (delete directory for standalone)
        reinit_and_delete()



"""
Module for analysing synaptic pre and post code for synapse order independence.
"""

from brian2.core.functions import Function
from brian2.core.variables import Constant
from brian2.utils.stringtools import get_identifiers

__all__ = ["OrderDependenceError", "check_for_order_independence"]


class OrderDependenceError(Exception):
    pass


def check_for_order_independence(statements, variables, indices):
    """
    Check that the sequence of statements doesn't depend on the order in which the indices are iterated through.
    """
    # Remove stateless functions from variables (only bother with ones that are used)
    all_used_vars = set()
    for statement in statements:
        all_used_vars.update(get_identifiers(statement.expr))
    variables = variables.copy()
    for var in set(variables.keys()).intersection(all_used_vars):
        val = variables[var]
        if isinstance(val, Function):
            if val.stateless:
                del variables[var]
            else:
                raise OrderDependenceError(
                    "Function %s may have internal state, "
                    "which can lead to order dependence." % var
                )
    all_variables = [v for v in variables if not isinstance(variables[v], Constant)]
    # Main index variables are those whose index corresponds to the main index being iterated through. By
    # assumption/definition, these indices are unique, and any order-dependence cannot come from their values,
    # only from the values of the derived indices. In the most common case of Synapses, the main index would be
    # the synapse index, and the derived indices would be pre and postsynaptic indices (which can be repeated).
    unique_index = lambda v: (
        indices[v] != "0" and getattr(variables[indices[v]], "unique", False)
    )
    main_index_variables = {
        v for v in all_variables if indices[v] == "_idx" or unique_index(v)
    }
    different_index_variables = set(all_variables) - main_index_variables

    # At the start, we assume all the different/derived index variables are permutation independent and we continue
    # to scan through the list of statements checking whether or not permutation-dependence has been introduced
    # until the permutation_independent set has stopped changing.
    permutation_independent = list(different_index_variables)
    permutation_dependent_aux_vars = set()
    changed_permutation_independent = True
    for statement in statements:
        if statement.op == ":=" and statement.var not in all_variables:
            main_index_variables.add(statement.var)
            all_variables.append(statement.var)

    while changed_permutation_independent:
        changed_permutation_independent = False
        for statement in statements:
            vars_in_expr = get_identifiers(statement.expr).intersection(all_variables)
            # any time a statement involves a LHS and RHS which only depend on itself, this doesn't change anything
            if {statement.var} == vars_in_expr:
                continue
            nonsyn_vars_in_expr = vars_in_expr.intersection(different_index_variables)
            permdep = any(
                var not in permutation_independent for var in nonsyn_vars_in_expr
            )
            permdep = permdep or any(
                var in permutation_dependent_aux_vars for var in vars_in_expr
            )
            if statement.op == ":=":  # auxiliary variable created
                if permdep:
                    if statement.var not in permutation_dependent_aux_vars:
                        permutation_dependent_aux_vars.add(statement.var)
                        changed_permutation_independent = True
                continue
            elif statement.var in main_index_variables:
                if permdep:
                    raise OrderDependenceError()
            elif statement.var in different_index_variables:
                if statement.op in ("+=", "*=", "-=", "/="):
                    if permdep:
                        raise OrderDependenceError()
                    if statement.var in permutation_independent:
                        permutation_independent.remove(statement.var)
                        changed_permutation_independent = True
                elif statement.op == "=":
                    otheridx = [
                        v
                        for v in variables
                        if indices[v] not in (indices[statement.var], "_idx", "0")
                    ]
                    if any(var in otheridx for var in vars_in_expr):
                        raise OrderDependenceError()
                    if permdep:
                        raise OrderDependenceError()
                    if any(var in main_index_variables for var in vars_in_expr):
                        raise OrderDependenceError()
                else:
                    raise OrderDependenceError()
            else:
                raise AssertionError("Should never get here...")



#!/usr/bin/env python
"""
Cochlear neuron model of Rothman & Manis
----------------------------------------
Rothman JS, Manis PB (2003) The roles potassium currents play in
regulating the electrical activity of ventral cochlear nucleus neurons.
J Neurophysiol 89:3097-113.

All model types differ only by the maximal conductances.

Adapted from their Neuron implementation by Romain Brette
"""
from brian2 import *

#defaultclock.dt=0.025*ms # for better precision

'''
Simulation parameters: choose current amplitude and neuron type
(from type1c, type1t, type12, type 21, type2, type2o)
'''
neuron_type = 'type1c'
Ipulse = 250*pA

C = 12*pF
Eh = -43*mV
EK = -70*mV  # -77*mV in mod file
El = -65*mV
ENa = 50*mV
nf = 0.85  # proportion of n vs p kinetics
zss = 0.5  # steady state inactivation of glt
temp = 22.  # temperature in degree celcius
q10 = 3. ** ((temp - 22) / 10.)
# hcno current (octopus cell)
frac = 0.0
qt = 4.5 ** ((temp - 33.) / 10.)

# Maximal conductances of different cell types in nS
maximal_conductances = dict(
type1c=(1000, 150, 0, 0, 0.5, 0, 2),
type1t=(1000, 80, 0, 65, 0.5, 0, 2),
type12=(1000, 150, 20, 0, 2, 0, 2),
type21=(1000, 150, 35, 0, 3.5, 0, 2),
type2=(1000, 150, 200, 0, 20, 0, 2),
type2o=(1000, 150, 600, 0, 0, 40, 2) # octopus cell
)
gnabar, gkhtbar, gkltbar, gkabar, ghbar, gbarno, gl = [x * nS for x in maximal_conductances[neuron_type]]

# Classical Na channel
eqs_na = """
ina = gnabar*m**3*h*(ENa-v) : amp
dm/dt=q10*(minf-m)/mtau : 1
dh/dt=q10*(hinf-h)/htau : 1
minf = 1./(1+exp(-(vu + 38.) / 7.)) : 1
hinf = 1./(1+exp((vu + 65.) / 6.)) : 1
mtau =  ((10. / (5*exp((vu+60.) / 18.) + 36.*exp(-(vu+60.) / 25.))) + 0.04)*ms : second
htau =  ((100. / (7*exp((vu+60.) / 11.) + 10.*exp(-(vu+60.) / 25.))) + 0.6)*ms : second
"""

# KHT channel (delayed-rectifier K+)
eqs_kht = """
ikht = gkhtbar*(nf*n**2 + (1-nf)*p)*(EK-v) : amp
dn/dt=q10*(ninf-n)/ntau : 1
dp/dt=q10*(pinf-p)/ptau : 1
ninf =   (1 + exp(-(vu + 15) / 5.))**-0.5 : 1
pinf =  1. / (1 + exp(-(vu + 23) / 6.)) : 1
ntau =  ((100. / (11*exp((vu+60) / 24.) + 21*exp(-(vu+60) / 23.))) + 0.7)*ms : second
ptau = ((100. / (4*exp((vu+60) / 32.) + 5*exp(-(vu+60) / 22.))) + 5)*ms : second
"""

# Ih channel (subthreshold adaptive, non-inactivating)
eqs_ih = """
ih = ghbar*r*(Eh-v) : amp
dr/dt=q10*(rinf-r)/rtau : 1
rinf = 1. / (1+exp((vu + 76.) / 7.)) : 1
rtau = ((100000. / (237.*exp((vu+60.) / 12.) + 17.*exp(-(vu+60.) / 14.))) + 25.)*ms : second
"""

# KLT channel (low threshold K+)
eqs_klt = """
iklt = gkltbar*w**4*z*(EK-v) : amp
dw/dt=q10*(winf-w)/wtau : 1
dz/dt=q10*(zinf-z)/ztau : 1
winf = (1. / (1 + exp(-(vu + 48.) / 6.)))**0.25 : 1
zinf = zss + ((1.-zss) / (1 + exp((vu + 71.) / 10.))) : 1
wtau = ((100. / (6.*exp((vu+60.) / 6.) + 16.*exp(-(vu+60.) / 45.))) + 1.5)*ms : second
ztau = ((1000. / (exp((vu+60.) / 20.) + exp(-(vu+60.) / 8.))) + 50)*ms : second
"""

# Ka channel (transient K+)
eqs_ka = """
ika = gkabar*a**4*b*c*(EK-v): amp
da/dt=q10*(ainf-a)/atau : 1
db/dt=q10*(binf-b)/btau : 1
dc/dt=q10*(cinf-c)/ctau : 1
ainf = (1. / (1 + exp(-(vu + 31) / 6.)))**0.25 : 1
binf = 1. / (1 + exp((vu + 66) / 7.))**0.5 : 1
cinf = 1. / (1 + exp((vu + 66) / 7.))**0.5 : 1
atau =  ((100. / (7*exp((vu+60) / 14.) + 29*exp(-(vu+60) / 24.))) + 0.1)*ms : second
btau =  ((1000. / (14*exp((vu+60) / 27.) + 29*exp(-(vu+60) / 24.))) + 1)*ms : second
ctau = ((90. / (1 + exp((-66-vu) / 17.))) + 10)*ms : second
"""

# Leak
eqs_leak = """
ileak = gl*(El-v) : amp
"""

# h current for octopus cells
eqs_hcno = """
ihcno = gbarno*(h1*frac + h2*(1-frac))*(Eh-v) : amp
dh1/dt=(hinfno-h1)/tau1 : 1
dh2/dt=(hinfno-h2)/tau2 : 1
hinfno = 1./(1+exp((vu+66.)/7.)) : 1
tau1 = bet1/(qt*0.008*(1+alp1))*ms : second
tau2 = bet2/(qt*0.0029*(1+alp2))*ms : second
alp1 = exp(1e-3*3*(vu+50)*9.648e4/(8.315*(273.16+temp))) : 1
bet1 = exp(1e-3*3*0.3*(vu+50)*9.648e4/(8.315*(273.16+temp))) : 1 
alp2 = exp(1e-3*3*(vu+84)*9.648e4/(8.315*(273.16+temp))) : 1
bet2 = exp(1e-3*3*0.6*(vu+84)*9.648e4/(8.315*(273.16+temp))) : 1
"""

eqs = """
dv/dt = (ileak + ina + ikht + iklt + ika + ih + ihcno + I)/C : volt
vu = v/mV : 1  # unitless v
I : amp
"""
eqs += eqs_leak + eqs_ka + eqs_na + eqs_ih + eqs_klt + eqs_kht + eqs_hcno

neuron = NeuronGroup(1, eqs, method='exponential_euler')
neuron.v = El

run(50*ms, report='text')  # Go to rest

M = StateMonitor(neuron, 'v', record=0)
neuron.I = Ipulse

run(100*ms, report='text')

plot(M.t / ms, M[0].v / mV)
xlabel('t (ms)')
ylabel('v (mV)')
show()
